(*
 * Builtin shell operations.
 *
 * \begin{doc}
 * \section{Shell functions}
 * \end{doc}
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf
open Lm_location

open Omake_ir
open Omake_env
open Omake_node
open Omake_value
open Omake_state
open Omake_symbol
open Omake_builtin_util
open Omake_shell_type

module Pos = MakePos (struct let name = "Omake_builtin_io" end)
open Pos

(*
 * The string should be a job identifier.
 *)
let pid_of_string pos s =
   try int_of_string s with
      Invalid_argument _
    | Failure _ ->
         let pos = string_pos "pid_of_string" pos in
            raise (OmakeException (pos, StringStringError ("not a process identifier", s)))

(*
 * Signal numbers.
 *)
let signal_of_string s =
   match String.uppercase s with
      "ABRT" -> SigAbrt
    | "ALRM" -> SigAlrm
    | "HUP"  -> SigHup
    | "ILL"  -> SigIll
    | "INT"  -> SigInt
    | "KILL" -> SigKill
    | "QUIT" -> SigQuit
    | "SEGV" -> SigSegv
    | "TERM" -> SigTerm
    | "USR1" -> SigUsr1
    | "USR2" -> SigUsr2
    | "CHLD" -> SigChld
    | "STOP" -> SigStop
    | "TSTP" -> SigTstp
    | "TTIN" -> SigTtin
    | "TTOU" -> SigTtou
    | "VTALRM" -> SigVTAlrm
    | "PROF" -> SigProf
    | _ ->
         SigNum (int_of_string s)

(*
 * Print some text.
 *
 * \begin{doc}
 * \subsection{echo}
 *
 * The \verb+echo+ function prints a string.
 *
 * \verb+$(echo <args>)+
 * \verb+echo <args>+
 * \end{doc}
 *)
let echo_fun venv pos loc args =
   let pos = string_pos "echo" pos in
   let args = List.map (strings_of_value venv pos) args in
   let args = List.flatten args in
   let outx = channel_of_value venv pos (venv_find_var venv ScopePublic pos loc stdout_sym) in
   let rec echo args =
      match args with
         [arg] ->
            Lm_channel.output_string outx arg
       | arg :: args ->
            Lm_channel.output_string outx arg;
            Lm_channel.output_char outx ' ';
            echo args
       | [] ->
            ()
   in
      echo args;
      Lm_channel.output_char outx '\n';
      Lm_channel.flush outx;
      ValNone

(*
 * \begin{doc}
 * \subsection{jobs}
 *
 * The \verb+jobs+ function prints a list of jobs.
 *
 * \verb+jobs+
 * \end{doc}
 *)
let jobs_fun venv pos loc args =
   let pos = string_pos "jobs" pos in
      Omake_shell_job.jobs venv;
      ValNone

(*
 * \begin{doc}
 * \subsection{cd}
 *
 * The \verb+cd+ function changes the current directory.
 *
 * \verb+cd <dir>+.
 * \end{doc}
 *)
let cd_fun venv pos loc args =
   let pos = string_pos "cd" pos in
      match args with
         [arg] ->
            let dir =
               match values_of_value venv pos arg with
                  [dir] ->
                     dir_of_value venv pos dir
                | [] ->
                     venv_intern_dir venv home_dir
                | _ ->
                     raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))
            in
            let venv = venv_chdir_dir venv dir in

            (* Check the directory for sanity *)
            let name = Dir.fullname dir in
            let result =
               try
                  let stat = Unix.stat name in
                     if stat.Unix.st_kind <> Unix.S_DIR then
                        raise Not_found;
                     Unix.access name [Unix.X_OK];
                     ValEnv (venv, ExportAll)
               with
                  Not_found ->
                     let relname = Dir.name (venv_dir venv) dir in
                        eprintf "%s: No such directory@." relname;
                        ValNone
                | Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               result
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(*
 * \begin{doc}
 * \subsection{bg}
 *
 * The \verb+bg+ function places a job in the background.
 *
 * \verb+bg <pid...>+
 * \end{doc}
 *)
let jobs_iter_fun f venv pos loc args =
   let pos = string_pos "iter" pos in
      match args with
         [arg] ->
            let pids = strings_of_value venv pos arg in
            let pids = List.map (pid_of_string pos) pids in
               List.iter (f venv pos) pids
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

let bg_fun venv pos loc args =
   let pos = string_pos "bg" pos in
      jobs_iter_fun Omake_shell_job.bg venv pos loc args;
      ValNone

(*
 * \begin{doc}
 * \subsection{fg}
 *
 * The \verb+fg+ function brings a job to the foreground.
 *
 * \verb+fg <pid...>+
 * \end{doc}
 *)
let fg_fun venv pos loc args =
   let pos = string_pos "fg" pos in
      jobs_iter_fun Omake_shell_job.fg venv pos loc args;
      ValNone

(*
 * \begin{doc}
 * \subsection{stop}
 *
 * The \verb+stop+ function suspends a job.
 *
 * \verb+stop <pid...>+
 * \end{doc}
 *)
let stop_fun venv pos loc args =
   let pos = string_pos "stop" pos in
      jobs_iter_fun Omake_shell_job.stop venv pos loc args;
      ValNone

(*
 * \begin{doc}
 * \subsection{wait}
 *
 * The \verb+wait+ function waits for a job to finish.
 * If no process identifiers are given, the shell waits for
 * all jobs to complete.
 *
 * \verb+wait <pid...>+
 * \end{doc}
 *)
let wait_fun venv pos loc args =
   let pos = string_pos "wait" pos in
      jobs_iter_fun Omake_shell_job.wait venv pos loc args;
      ValNone

(*
 * \begin{doc}
 * \subsection{kill}
 *
 * The \verb+kill+ function signals a job.
 *
 * \verb+kill [signal] <pid...>+
 * \end{doc}
 *)
let kill_fun venv pos loc args =
   let pos = string_pos "kill" pos in
      match args with
         [signal; arg] ->
            let signal = signal_of_string (string_of_value venv pos signal) in
            let args = strings_of_value venv pos arg in
            let args = List.map (pid_of_string pos) args in
               List.iter (fun pid -> Omake_shell_job.kill venv pos pid signal) args;
               ValNone
       | _ ->
            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))

(************************************************************************
 * Tables.
 *)

(*
 * Table of variables.
 *)
let builtin_vars =
   []

(*
 * Table of built-in functions.
 *)
let builtin_funs =
   [true, "echo",                  echo_fun,                 ArityAny;
    true, "jobs",                  jobs_fun,                 ArityExact 0;
    true, "bg",                    bg_fun,                   ArityExact 1;
    true, "fg",                    fg_fun,                   ArityExact 1;
    true, "stop",                  stop_fun,                 ArityExact 1;
    true, "wait",                  wait_fun,                 ArityExact 1;
    true, "cd",                    cd_fun,                   ArityExact 1;
    true, "kill",                  kill_fun,                 ArityExact 1]

(*
 * Builtin rules.
 *)
let phony_targets =
   []

let builtin_rules =
   []

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
