(*
 * Keep a cache that determines whether we _really_ need to run
 * a command to build a file.
 *
 * The basic idea is this: for each file, remember the command
 * that built it, its Digest, and the Digest of the files it
 * depends on.  If we want to build the node again, and all
 * the parts are the same, then we don't actually need to build
 * it.
 *
 * The same principle works for scanners.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf

open Lm_debug

open Omake_node
open Omake_command
open Omake_node_sig
open Omake_cache_type

let debug_cache =
   create_debug (**)
      { debug_name = "cache";
        debug_description = "Display debugging information for the cache";
        debug_value = false
      }

(*
 * The cache type.
 *)
type t = cache

(*
 * Memo functions for the cache.
 *)
let scanner_fun = 0
let rule_fun    = 1
let env_fun     = 2
let include_fun = 3
let include_scanner_fun = 4

let env_target  = Node.phony_global ".ENV"

(************************************************************************
 * Printing.
 *)

(*
 * Print a digest.
 *)
let pp_print_digest buf digest =
   let s =
      match digest with
         Some digest ->
            Lm_string_util.hexify digest
       | None ->
            "<none>"
   in
      pp_print_string buf s

(*
 * Print a digest table.
 *)
let pp_print_node_digest_table buf deps =
   NodeTable.iter (fun node digest ->
         fprintf buf "@ %a = %a" (**)
            pp_print_node node
            pp_print_digest digest) deps

(*
 * Print a result.
 *)
let pp_print_memo_result pp_print_result buf result =
   match result with
      MemoNone ->
         ()
    | MemoFailed code ->
         fprintf buf "@ failed(%d)" code
    | MemoResult result ->
         pp_print_result buf result

(*
 * Print a memo.
 *)
let pp_print_memo buf memo =
   let { memo_index = index;
         memo_targets = targets;
         memo_deps = deps;
         memo_result = result;
         memo_commands = commands
       } = memo
   in
      fprintf buf "@[<hv 0>@[<hv 3>memo {@ index = %d;@ @[<b 3>targets =%a@];@ @[<b 3>deps =%a@]@ @[<hv 3>commands =%a@]@]@ }@]" (**)
         index
         pp_print_node_digest_table targets
         pp_print_node_digest_table deps
         pp_print_command_lines commands

(************************************************************************
 * Persistence.
 *)

(*
 * Create the cache.  We save the cache in a file, so try to
 * load it from that file.
 *)
let magic_number = Omake_magic.cache_magic
let input_magic  = Omake_magic.input_magic
let output_magic = Omake_magic.output_magic

(*
 * Create a new cache.
 *)
let create () =
   { cache_nodes        = NodeTable.empty;
     cache_info         = [||];
     cache_stat_count   = 0;
     cache_digest_count = 0
   }

let stats { cache_stat_count = stat_count;
            cache_digest_count = digest_count
    } =
   stat_count, digest_count

(*
 * Clear one of the tables.
 *)
let clear cache index =
   let info = cache.cache_info in
      if index < Array.length info then
         let info = info.(index) in
            info.cache_memos <- NodeTable.empty;
            info.cache_index <- IndexMTable.empty

(*
 * When the file is saved, remove all stat information,
 * as well as any files that don't exist.
 *)
let save_of_cache cache =
   let { cache_nodes = cache_nodes;
         cache_info  = cache_info
       } = cache
   in
   let nodes = Node.marshal_base () in
   let cache_nodes =
      NodeTable.fold (fun nodes target nmemo ->
            match nmemo.nmemo_stats with
               FreshStats stats
             | OldStats stats ->
                  let nmemo = { nmemo with nmemo_stats = OldStats stats } in
                     NodeTable.add nodes target nmemo
             | NoStats ->
                  nodes) NodeTable.empty cache_nodes
   in
   let cache_info = Array.map (fun info -> info.cache_memos) cache_info in
      { save_nodes = nodes;
        save_cache_nodes = cache_nodes;
        save_cache_info  = cache_info
      }

(*
 * Rebuild the index from the memos.
 *)
let create_index memos =
   NodeTable.fold (fun (memos, index) target memo ->
         let { memo_result = result;
               memo_index = hash
             } = memo
         in
            match result with
               MemoNone
             | MemoResult _ ->
                  let index = IndexMTable.add index hash target in
                     memos, index
             | MemoFailed _ ->
                  let memos = NodeTable.remove memos target in
                     memos, index) (memos, IndexMTable.empty) memos

(*
 * Rebuild the cache from the saved version.
 *)
let cache_of_save save =
   let { save_nodes = nodes;
         save_cache_nodes = cache_nodes;
         save_cache_info  = cache_info
       } = save
   in
   let cache_info =
      Array.map (fun memos ->
            let memos, index = create_index memos in
               { cache_memos = memos;
                 cache_index = index
               }) cache_info
   in
      Node.unmarshal_base nodes;
      { cache_nodes = cache_nodes;
        cache_info  = cache_info;
        cache_stat_count = 0;
        cache_digest_count = 0
      }

(*
 * Load the old cache from a file.
 *)
let from_channel inx =
   let magic = input_magic inx in
   let _ =
      if magic <> magic_number then
         raise (Sys_error "bad magic number")
   in
   let save = Marshal.from_channel inx in
      cache_of_save save

(*
 * Save the cache to the file.
 *)
let to_channel outx cache =
   output_magic outx magic_number;
   Marshal.to_channel outx (save_of_cache cache) []

(************************************************************************
 * Stat.
 *)

(*
 * Reset the information about a node (probably because
 * we ran a command that changed it).
 *)
let reset cache node =
   let nodes = cache.cache_nodes in
   let nodes =
      try
         NodeTable.filter_remove nodes node (fun nmemo ->
               match nmemo.nmemo_stats with
                  FreshStats stats
                | OldStats stats ->
                     Some { nmemo with nmemo_stats = OldStats stats }
                | NoStats ->
                     None)
      with
         Not_found ->
            nodes
   in
      cache.cache_nodes <- nodes

let reset_set cache nodes =
   NodeSet.iter (reset cache) nodes

let reset_table cache nodes =
   NodeTable.iter (fun node _ -> reset cache node) nodes

(*
 * Test whether stats are equal enough that we think the
 * file is up-to-date.
 *)
let stats_equal stat1 stat2 =
   let { Unix.st_ino  = ino1;
         Unix.st_kind = kind1;
         Unix.st_size = size1;
         Unix.st_mtime = mtime1
       } = stat1
   in
   let { Unix.st_ino  = ino2;
         Unix.st_kind = kind2;
         Unix.st_size = size2;
         Unix.st_mtime = mtime2
       } = stat2
   in
      ino1 = ino2 && kind1 = kind2 && size1 = size2 && mtime1 = mtime2

(*
 * Stat a file.
 *)
let stat_file cache node =
   let nodes = cache.cache_nodes in
   let stats =
      try Some (NodeTable.find nodes node) with
         Not_found ->
            None
   in
      match stats with
         Some { nmemo_stats = FreshStats _;
                nmemo_digest = digest
         }
       | Some { nmemo_stats = NoStats;
                nmemo_digest = digest
         } ->
            (*
             * We've recently computed the digest for this file.
             *)
            digest
       | Some ({ nmemo_stats = OldStats stats } as nmemo) ->
            (*
             * We have no recent record of this file.
             * Get current stats.  If they match, then
             * use current digest; otherwise recompute.
             *)
            let name = Node.fullname node in
            let nmemo =
               try
                  cache.cache_stat_count <- succ cache.cache_stat_count;
                  let stats' = Unix.stat name in
                     if stats_equal stats' stats then
                        { nmemo with nmemo_stats = FreshStats stats' }
                     else if stats'.Unix.st_kind = Unix.S_DIR then
                        { nmemo_stats = FreshStats stats';
                          nmemo_digest = squash_stat
                        }
                     else
                        let digest = Digest.file name in
                           cache.cache_digest_count <- succ cache.cache_digest_count;
                           { nmemo_stats = FreshStats stats';
                             nmemo_digest = Some digest
                           }
               with
                  Unix.Unix_error _
                | Sys_error _ ->
                     { nmemo_stats = NoStats;
                       nmemo_digest = None
                     }
            in
               cache.cache_nodes <- NodeTable.add nodes node nmemo;
               nmemo.nmemo_digest

       | None ->
            (*
             * We've never seen this file before.
             * Get complete stats, including a digest.
             *)
            let name = Node.fullname node in
            let nmemo =
               try
                  cache.cache_stat_count <- succ cache.cache_stat_count;
                  let stats = Unix.stat name in
                     if stats.Unix.st_kind = Unix.S_DIR then
                        { nmemo_stats = FreshStats stats;
                          nmemo_digest = squash_stat
                        }
                     else
                        let digest = Digest.file name in
                           cache.cache_digest_count <- succ cache.cache_digest_count;
                           { nmemo_stats = FreshStats stats;
                             nmemo_digest = Some digest
                           }
               with
                  Unix.Unix_error _
                | Sys_error _ ->
                     { nmemo_stats = NoStats;
                       nmemo_digest = None
                     }
            in
               cache.cache_nodes <- NodeTable.add nodes node nmemo;
               nmemo.nmemo_digest

(*
 * Return stat info.
 *)
let rec stat_unix cache node =
   match (NodeTable.find cache.cache_nodes node).nmemo_stats with
      FreshStats stats ->
         stats
    | NoStats ->
         raise Not_found
    | _ ->
         let _ = stat_file cache node in
            stat_unix cache node

(*
 * Check if a file is a directory.
 *)
let is_dir cache node =
   try (stat_unix cache node).Unix.st_kind = Unix.S_DIR with
      Not_found ->
         false

(*
 * Check if the node is phony first.
 *)
let stat cache node =
   let core = Node.core node in
      match Node.kind node with
         NodePhony ->
            None
       | NodeOptional
       | NodeNormal ->
            stat_file cache core
       | NodeExists ->
            (match stat_file cache core with
                Some _ ->
                   squash_stat
              | None ->
                   None)

(*
 * Turn a set into a table of stat info.
 *)
let stat_set cache nodes =
   NodeSet.fold (fun table node ->
         NodeTable.add table node (stat cache node)) NodeTable.empty nodes

let stat_table cache nodes =
   NodeTable.mapi (fun node _ -> stat cache node) nodes

(*
 * Force a stat.
 *)
let force_stat cache node =
   reset cache node;
   stat cache node

let force_stat_set cache nodes =
   NodeSet.fold (fun table node ->
         NodeTable.add table node (force_stat cache node)) NodeTable.empty nodes

let force_stat_table cache nodes =
   NodeTable.mapi (fun node _ -> force_stat cache node) nodes

(*
 * Check if the stat changed.
 *)
let stat_changed cache node =
   let old_digest =
      try (NodeTable.find cache.cache_nodes node).nmemo_digest with
         Not_found ->
            None
   in
   let new_digest = force_stat cache node in
      new_digest <> old_digest

(*
 * Check if a file exists.
 *)
let exists cache node =
   match stat cache node with
      Some _ -> true
    | None -> Node.always_exists node

(************************************************************************
 * Adding to the cache.
 *)

(*
 * Hash a set of deps and commands.
 *)
let hash_index deps commands =
   let index =
      NodeSet.fold (fun index node ->
            index lxor (index lsl 4) lxor (index lsr 4) lxor (Node.hash node)) 0 deps
   in
   let index =
      List.fold_left (fun index command ->
            match command.command_inst with
               CommandArgv s ->
                  index lxor (index lsl 4) lxor (index lsr 4) lxor (Hashtbl.hash s)
             | CommandEval _ ->
                  index lxor (index lsl 4) lxor (index lsr 4)) index commands
   in
      index land 0x3fffffff

(*
 * Expand the cache_info if necessary.
 *)
let get_info cache key =
   let cache_info = cache.cache_info in
   let len = Array.length cache_info in
   let cache_info =
      if key >= Array.length cache_info then
         begin
            let cache_info' =
               Array.init (succ key) (fun _ ->
                     { cache_memos = NodeTable.empty;
                       cache_index = IndexMTable.empty
                     })
            in
               Array.blit cache_info 0 cache_info' 0 len;
               cache.cache_info <- cache_info';
               cache_info'
         end
      else
         cache_info
   in
      cache_info.(key)

(*
 * Add a command.
 *)
let add cache key targets deps commands result =
   let target = NodeSet.choose targets in
   let commands = squash_command_lines commands in
   let index = hash_index deps commands in
   let memo =
      { memo_index    = index;
        memo_targets  = stat_set cache targets;
        memo_deps     = stat_set cache deps;
        memo_result   = result;
        memo_commands = commands
      }
   in
   let info = get_info cache key in
      info.cache_memos <- NodeTable.add info.cache_memos target memo;
      info.cache_index <- IndexMTable.add info.cache_index index target

(*
 * Check the target digest.
 *)
let targets_equal cache targets =
   NodeTable.forall (fun target digest ->
         stat cache target = digest) targets

(*
 * Check if deps are the same.
 * This returns true if the deps are equal.
 * If not, it either returns false, or raises Not_found.
 *)
let deps_equal cache deps1 deps2 =
   let count1 = NodeSet.cardinal deps1 in
   let count2 = NodeTable.cardinal deps2 in
      (count1 = count2) && NodeSet.for_all (fun dep1 ->
            let digest1 = stat cache dep1 in
            let digest2 = NodeTable.find deps2 dep1 in
               digest1 = digest2) deps1

(*
 * Find a memo from the deps and commands.
 *)
let find_memo cache key deps commands =
   let { cache_memos = memos;
         cache_index = index
       } = get_info cache key
   in
   let hash = hash_index deps commands in
   let rec search = function
      target :: targets ->
         let memo = NodeTable.find memos target in
         let { memo_index = hash';
               memo_deps  = deps';
               memo_commands = commands'
             } = memo
         in
            if (hash' = hash)
               && (deps_equal cache deps deps')
               && (commands_are_equal commands commands')
            then
               memo
            else
               search targets
    | [] ->
         raise Not_found
   in
      search (IndexMTable.find_all index hash)

(*
 * A memo has not changed if all the deps and the target
 * have the same digests.
 *)
let up_to_date cache key deps commands =
   try
      let memo = find_memo cache key deps commands in
         targets_equal cache memo.memo_targets
   with
      Not_found ->
         false

(*
 * A memo has not changed if all the deps and the target
 * have the same digests.
 *)
let find_result cache key deps commands =
   let memo = find_memo cache key deps commands in
   let { memo_targets = targets;
         memo_result  = result
       } = memo
   in
      if targets_equal cache targets then
         result
      else
         raise Not_found

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
