(*
 * Types used to represent commands and the cache.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_location

open Omake_ir
open Omake_node

(*
 * An index table.
 * The indexes are small enough that (-) is a correct ordering function.
 *)
module IndexCompare =
struct
   type t = int
   let compare = (-)
end

module IndexMTable = Lm_map.LmMakeList (IndexCompare)

(*
 * A command line is a string, together with come flags.
 *)
type command_flag =
   QuietFlag
 | AllowFailureFlag
 | AllowOutputFlag

(*
 * The command line has some flags,
 * and a string to be executed internally
 * or passed to the shell.
 *)
type 'exp poly_command_inst =
   CommandEval of 'exp
 | CommandArgv of string list

type ('venv, 'exp) poly_command_line =
   { command_loc    : loc;
     command_dir    : Dir.t;
     command_target : Node.t;
     command_flags  : command_flag list;
     command_venv   : 'venv;
     command_inst   : 'exp poly_command_inst
   }

type unit_command_inst = unit poly_command_inst
type unit_command_line = (unit, unit) poly_command_line

(*
 * File digest is an option, in case the file does not exist.
 *)
type digest = Digest.t option

(*
 * The result is a scanner table.
 *)
type scanner_result = NodeSet.t NodeTable.t

(*
 * The memo result is used only for the scanner,
 * whice produces a table of dependencies.
 *)
type memo_result =
   MemoNone
 | MemoFailed of int
 | MemoResult of scanner_result

(*
 * The memo is a record of what we have done during
 * a previous run of this program.
 *
 * The index is a hash over the deps and the commands.
 * The result is only used for the scanner.
 *)
type memo =
   { memo_index    : int;
     memo_targets  : digest NodeTable.t;
     memo_deps     : digest NodeTable.t;
     memo_commands : unit_command_line list;
     memo_result   : memo_result
   }

(*
 * Stats of a file.
 * FreshStats: we have taken the stats during this program run
 * OldStats: we took the stats during a prior run of this program
 * NoStats: we tried to take stats, but an error occurred (usually
 *    because the file did not exist).
 *)
type node_stats =
   FreshStats of Unix.stats
 | OldStats of Unix.stats
 | NoStats

(*
 * For each file, we keep the stats and the digest.
 *)
type node_memo =
   { nmemo_stats  : node_stats;
     nmemo_digest : digest
   }

(*
 * The memos are classified by functions.
 * Each function has a table of memos, and an index.
 *)
type key = int

type cache_info =
   { mutable cache_memos     : memo NodeTable.t;
     mutable cache_index     : Node.t IndexMTable.t
   }

(*
 * The cache remembers all the build commands we did last time.
 *)
type cache =
   { mutable cache_nodes        : node_memo NodeTable.t;
     mutable cache_info         : cache_info array;
     mutable cache_stat_count   : int;
     mutable cache_digest_count : int
   }

(*
 * The version of the cache that is saved in the file also
 * contains a table of all nodes.
 *)
type cache_save =
   { save_nodes              : Node.db;
     save_cache_nodes        : node_memo NodeTable.t;
     save_cache_info         : memo NodeTable.t array
   }

(*
 * Squash stat code.
 *)
let squash_stat = Some "squash"

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
