(*
 * The environment for evaluating programs.
 *
 * Scoping is primarily dynamic, but there are issues.
 * Consider the following code.
 *
 *    F(x) =
 *       A = $(x)
 *       G() =
 *          println(A=$(A))
 *       return($(G))
 *    G = $(F 1)
 *    G()
 *
 * With pure dynamic scope, the variable A is unbound in the
 * function call.
 *
 * The same problem occurs with nested objects.
 * Consider the following program.
 *
 *   A. =
 *      X = 1
 *      B. =
 *         F() =
 *            println(X=$(X))
 *
 * The inner object B refers to the variable X defined in
 * the outer object A.
 *
 * We expect the following code to print X=17:
 *
 *    X = 17
 *    A.B.F()
 *
 * What about calling F without defining a value for X?
 *
 *    A.B.F()
 *
 * There are two choices here:
 *    1. X is undefined
 *    2. X gets the value from A (so the program prints X=1)
 *
 * In general we address this by using both static and
 * dynamic scope.  Dynamic scoping takes precedence,
 * but if a dynamic lookup fails, fall back to the
 * the static scope.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf

open Lm_debug
open Lm_symbol
open Lm_location
open Lm_string_set
open Lm_string_util

open Omake_ir
open Omake_wild
open Omake_node
open Omake_exec
open Omake_state
open Omake_lexer
open Omake_parser
open Omake_symbol
open Omake_ir_print
open Omake_node_sig
open Omake_print_util
open Omake_cache_type
open Omake_options_type

(*
 * Debugging.
 *)
let debug_scanner =
   create_debug (**)
      { debug_name = "scanner";
        debug_description = "Display debugging information for scanner selection";
        debug_value = false
      }

(*
 * Debugging.
 *)
let debug_implicit =
   create_debug (**)
      { debug_name = "implicit";
        debug_description = "Display debugging information for implicit rule selection";
        debug_value = false
      }

(*
 * A source is either
 *   1. A wild string
 *   2. A node
 *   3. An optional source
 *   4. A squashed source
 *)
type source_core =
   SourceWild of wild
 | SourceNode of Node.t

type 'a source = node_kind * 'a

(*
 * IoChannels.
 *)
type channel_mode = Lm_channel.mode =
   InChannel
 | OutChannel
 | InOutChannel

type channel_data =
   ChannelDelayed
 | ChannelClosed
 | ChannelValue of Lm_channel.t

type prim_channel =
   { channel_id           : int;
     mutable channel_data : channel_data
   }

(*
 * Possible values.
 * Someday we may want to include rules and functions.
 * For the function, the obj is the static scope.
 *)
type value =
   ValNone
 | ValInt         of int
 | ValFloat       of float
 | ValSequence    of value list
 | ValArray       of value list
 | ValString      of string
 | ValData        of string
 | ValQuote       of value list
 | ValApply       of loc * scope_kind * var * value list
 | ValSuperApply  of loc * scope_kind * var * var * value list
 | ValMethodApply of loc * scope_kind * var list * value list
 | ValFun         of arity * env * var list * exp
 | ValFunValue    of arity * env * var list * value
 | ValPrim        of arity * bool * prim_fun
 | ValRules       of erule list
 | ValNode        of Node.t
 | ValDir         of Dir.t
 | ValEnv         of venv * export
 | ValBody        of exp
 | ValObject      of obj
 | ValMap         of obj
 | ValChannel     of channel_mode * prim_channel
 | ValClass       of obj SymbolTable.t
 | ValCases       of (var * value * value) list
 | ValOther       of value_other

and value_other =
   ValLexer       of Lexer.t
 | ValParser      of Parser.t
 | ValLocation    of loc
 | ValPosition    of pos

(*
 * Export kinds.
 *    ExportFile: this is the result of file evaluation
 *    ExportSymbols syms: export just the symbols in the list
 *    ExportAll: export the entire environment.
 *    ExportValue v: just like Exportall, but return the value too
 *)
and export =
   ExportFile
 | ExportAll
 | ExportSymbols of symbol list
 | ExportValue of value

(*
 * Primitives are option refs.
 * We do this so that we can marshal these values.
 * Just before marshaling, all the options are set to None.
 *)
and prim_fun_data = (venv -> pos -> loc -> value list -> value) option

and prim_fun =
   { fun_id           : symbol;
     mutable fun_data : prim_fun_data
   }

(*
 * An object is just an environment.
 *)
and obj = value SymbolTable.t
and env = value SymbolTable.t

(*
 * Command lists have source arguments.
 *)
and command =
   CommandSection of value * exp
 | CommandValue of value

and command_info =
   { command_env     : venv;
     command_sources : Node.t list;
     command_body    : command list
   }

(*
 * In an implicit rule, we compile the targets/sources
 * to wild patterns.
 *)
and irule =
   { irule_loc      : loc;
     irule_patterns : wild list;
     irule_effects  : source_core list;
     irule_sources  : source_core source list;
     irule_body     : command list
   }

(*
 * An implicit dependency.
 *)
and inrule =
   { inrule_loc      : loc;
     inrule_patterns : wild list;
     inrule_sources  : source_core source list
   }

(*
 * The internal representation of a scanner rule.
 *)
and srule =
   { srule_loc      : loc;
     srule_patterns : wild list;
     srule_sources  : source_core source list;
     srule_body     : command list
   }

(*
 * Scanner dependency.
 *)
and snrule =
   { snrule_patterns : wild list;
     snrule_sources  : source_core source list
   }

(*
 * The external representation of a scanner rule.
 *)
and scanner =
   { scanner_loc      : loc;
     scanner_target   : Node.t;
     scanner_sources  : NodeSet.t;
     scanner_match    : string;
     scanner_body     : command_info list
   }

(*
 * Explicit rules.
 *)
and erule =
   { rule_loc        : loc;
     rule_env        : venv;
     rule_target     : Node.t;
     rule_effects    : NodeSet.t;
     rule_sources    : NodeSet.t;
     rule_match      : string option;
     rule_multiple   : bool;
     rule_commands   : command_info list
   }

(*
 * Info about explicit rules.
 * Keep track of all the possible explicit targets.
 *)
and erule_info =
   { explicit_targets     : erule NodeTable.t;
     explicit_directories : venv DirTable.t;
     explicit_deps        : NodeSet.t NodeTable.t
   }

(*
 * An ordering rule.
 * For now, this just defines an extra dependency
 * of the form:  patt1 -> patt2
 * This means that if a file depends on patt1,
 * then it also depends on patt2.
 *)
and orule =
   { orule_loc      : loc;
     orule_name     : symbol;
     orule_pattern  : wild;
     orule_sources  : source_core list
   }

and ordering_info = orule list

(*
 * The environment contains three scopes:
 *    1. The dynamic scope
 *    2. The current object
 *    3. The static scope
 * Lookup occurs in that order, unless the variables
 * have been defined otherwise.
 *
 * Each function has its own static scope.
 * The dynamic scope comes from the caller.
 *)
and venv =
   { venv_dynamic        : env;
     venv_this           : obj;
     venv_static         : env;
     venv_inner          : venv_inner
   }

and venv_inner =
   { venv_environ        : string SymbolTable.t;
     venv_dir            : Dir.t;
     venv_phony          : NodeSet.t;
     venv_no_scanner     : symbol;
     venv_scanner_deps   : snrule list;
     venv_scanner_rules  : srule list;
     venv_implicit_deps  : inrule list;
     venv_implicit_rules : irule list;
     venv_options        : omake_options;
     venv_globals        : venv_globals
   }

and venv_globals =
   { venv_cwd                      : Dir.t;
     venv_exec                     : exec;
     venv_cache                    : Omake_cache.t;
     mutable venv_files            : NodeSet.t;
     mutable venv_directories      : venv DirTable.t;
     mutable venv_explicit_rules   : erule list;
     mutable venv_explicit_targets : NodeSet.t;
     mutable venv_explicit_new     : erule list;
     mutable venv_ordering_rules   : orule list;
     mutable venv_orders           : StringSet.t;
     mutable venv_pervasives       : obj
   }

(*
 * Type of execution servers.
 *)
and pid =
   InternalPid of int
 | ExternalPid of int
 | ResultPid of int * value

and exec = (venv_command_line, pid, value) Exec.t

(*
 * Execution service.
 *)
and venv_command_inst = exp poly_command_inst
and venv_command_line = (venv, exp) poly_command_line

(*
 * Exceptions.
 *)
and item =
   Symbol        of symbol
 | String        of string
 | AstExp        of Omake_ast.exp
 | IrExp         of Omake_ir.exp
 | Location      of loc
 | Value         of value
 | Error         of omake_error

and pos = item Lm_position.pos

and omake_error =
   SyntaxError       of string
 | StringError       of string
 | StringStringError of string * string
 | StringNodeError   of string * Node.t
 | StringVarError    of string * var
 | StringIntError    of string * int
 | StringMethodError of string * var list
 | StringValueError  of string * value
 | LazyError         of (formatter -> unit)
 | UnboundVar        of var
 | UnboundFun        of var
 | UnboundMethod     of var list
 | ArityMismatch     of arity * int
 | NotImplemented    of string

(*
 * Error during translation.
 *)
exception OmakeException    of pos * omake_error
exception UncaughtException of pos * exn
exception RaiseException    of pos * obj

(*
 * Now the stuff that is really global, not saved in venv.
 *)
module IntCompare =
struct
   type t = int
   let compare = (-)
end;;

module IntTable = Lm_map.LmMake (IntCompare);;

type venv_runtime =
   { mutable venv_channel_index  : int;
     mutable venv_channels       : (prim_channel * channel_data) IntTable.t;
     mutable venv_primitives     : (prim_fun * prim_fun_data) SymbolTable.t
   }

let venv_runtime =
   { venv_channel_index  = 0;
     venv_channels       = IntTable.empty;
     venv_primitives     = SymbolTable.empty
   }

(************************************************************************
 * Special symbols.
 *)

let class_sym = Lm_symbol.add "$class"

(*
 * Basic utilities.
 *)
let venv_get_class obj =
   match
      try SymbolTable.find obj class_sym with
         Not_found ->
            ValNone
   with
      ValClass table ->
         table
    | _ ->
         SymbolTable.empty

(************************************************************************
 * Printing.
 *)
let pp_print_pos_ref = ref (fun _ _ -> ())

let pp_print_string_list buf sl =
   List.iter (fun s -> fprintf buf "@ %s" s) sl

let pp_print_node_list buf l =
   List.iter (fun s -> fprintf buf "@ %a" pp_print_node s) l

let pp_print_node_set buf set =
   NodeSet.iter (fun s -> fprintf buf "@ %a" pp_print_node s) set

let pp_print_wild_list buf wl =
   List.iter (fun w -> fprintf buf "@ %a" pp_print_wild w) wl

let pp_print_source buf (_, source) =
   match source with
      SourceWild wild ->
         pp_print_wild buf wild
    | SourceNode node ->
         pp_print_node buf node

let pp_print_source_list buf sources =
   List.iter (fun source -> fprintf buf "@ %a" pp_print_source source) sources

let rec pp_print_value buf v =
   match v with
      ValNone ->
         pp_print_string buf "<none>"
    | ValInt i ->
         fprintf buf "%d : Int" i
    | ValFloat x ->
         fprintf buf "%g : Float" x
    | ValData s ->
         fprintf buf "\"%s\" : String" (String.escaped s)
    | ValQuote vl ->
         fprintf buf "@[<v 3><string%a>@ : String@]" pp_print_value_list vl
    | ValString s ->
         fprintf buf "\"%s\" : Sequence" (String.escaped s)
    | ValSequence [v] ->
         pp_print_value buf v
    | ValSequence vl ->
         fprintf buf "@[<hv 3><sequence%a>@ : Sequence@]" pp_print_value_list vl
    | ValArray vl ->
         fprintf buf "@[<v 3><array%a>@ : Array@]" pp_print_value_list vl
    | ValApply (_, scope, f, args) ->
         fprintf buf "@[<hv 3>$(apply %a%a%a)@]" (**)
            pp_print_scope_kind scope
            pp_print_symbol f
            pp_print_value_list args
    | ValSuperApply (_, scope, super, f, args) ->
         fprintf buf "@[<hv 3>$(apply %a%a::%a%a)@]" (**)
            pp_print_scope_kind scope
            pp_print_symbol super
            pp_print_symbol f
            pp_print_value_list args
    | ValMethodApply (_, scope, vl, args) ->
         fprintf buf "@[<hv 3>$(%a%a%a)@]" (**)
            pp_print_scope_kind scope
            pp_print_method_name vl
            pp_print_value_list args
    | ValFun (arity, _, _, _)
    | ValFunValue (arity, _, _, _) ->
         fprintf buf "<fun %a>" pp_print_arity arity
    | ValPrim (arity, special, _) ->
         if special then
            fprintf buf "<special %a>" pp_print_arity arity
         else
            fprintf buf "<prim %a>" pp_print_arity arity
    | ValRules rules ->
         fprintf buf "<@[<hv 3>rules:";
         List.iter (fun erule -> fprintf buf "@ %a" pp_print_rule erule) rules;
         fprintf buf "@]>"
    | ValDir dir ->
         fprintf buf "%a : Dir" pp_print_dir dir
    | ValNode node ->
         fprintf buf "%a : File" pp_print_node node
    | ValEnv _ ->
         pp_print_string buf "<env>"
    | ValBody e ->
         fprintf buf "@[<v 0>%a@ : Body@]" pp_print_exp e
    | ValObject env ->
         let tags = venv_get_class env in
         let env = SymbolTable.remove env class_sym in
            fprintf buf "@[<v 3>@[<hv 3>class";
            SymbolTable.iter (fun v _ -> fprintf buf "@ %a" pp_print_symbol v) tags;
            fprintf buf "@]";
            SymbolTable.iter (fun v e -> fprintf buf "@ %a = %a" pp_print_symbol v pp_print_value e) env;
            fprintf buf "@]"
    | ValMap env ->
         let tags = venv_get_class env in
         let env = SymbolTable.remove env class_sym in
            fprintf buf "@[<v 3>@[<hv 3>map";
            SymbolTable.iter (fun v _ -> fprintf buf "@ %a" pp_print_symbol v) tags;
            fprintf buf "@]";
            SymbolTable.iter (fun v e -> fprintf buf "@ %a = %a" pp_print_symbol v pp_print_value e) env;
            fprintf buf "@]"
    | ValChannel (InChannel, _) ->
         fprintf buf "<channel> : InChannel"
    | ValChannel (OutChannel, _) ->
         fprintf buf "<channel> : OutChannel"
    | ValChannel (InOutChannel, _) ->
         fprintf buf "<channel> : InOutChannel"
    | ValClass c ->
         fprintf buf "@[<hv 3>class";
         SymbolTable.iter (fun v _ ->
               fprintf buf "@ %a" pp_print_symbol v) c;
         fprintf buf "@]"
    | ValCases cases ->
         fprintf buf "@[<hv 3>cases";
         List.iter (fun (v, e1, e2) ->
               fprintf buf "@[<hv 3>%a %a:@ %a@]" pp_print_symbol v pp_print_value e1 pp_print_value e2) cases;
         fprintf buf "@]"
    | ValOther (ValLexer _) ->
         fprintf buf "<lexer> : Lexer"
    | ValOther (ValParser _) ->
         fprintf buf "<parser> : Parser"
    | ValOther (ValLocation loc) ->
         fprintf buf "<location %a> : Location" pp_print_location loc
    | ValOther (ValPosition pos) ->
         fprintf buf "<position %a> : Position" !pp_print_pos_ref pos

and pp_print_value_list buf vl =
   List.iter (fun v -> fprintf buf "@ %a" pp_print_value v) vl

and pp_print_command buf command =
   match command with
      CommandSection (arg, e) ->
         fprintf buf "@[<hv 3>section %a@ %a@]" pp_print_value arg pp_print_exp e
    | CommandValue v ->
         pp_print_value buf v

and pp_print_commands buf commands =
   List.iter (fun command -> fprintf buf "@ %a" pp_print_command command) commands

and pp_print_command_info buf info =
   let { command_env     = venv;
         command_sources = sources;
         command_body    = commands
       } = info
   in
      fprintf buf "@[<hv 0>@[<hv 3>{@ command_dir = %a;@ @[<b 3>command_sources =%a@]@ @[<b 3>command_body =%a@]@]@ }@]" (**)
         pp_print_dir venv.venv_inner.venv_dir
         pp_print_node_list sources
         pp_print_commands commands

and pp_print_command_info_list buf infos =
   List.iter (fun info -> fprintf buf "@ %a" pp_print_command_info info) infos

and pp_print_rule buf erule =
   let { rule_target      = target;
         rule_effects     = effects;
         rule_sources     = sources;
         rule_commands    = commands
       } = erule
   in
      fprintf buf "@[<hv 0>@[<hv 3>rule {";
      fprintf buf "@ target = %a" pp_print_node target;
      fprintf buf "@ @[<b 3>effects =%a@]" pp_print_node_set effects;
      fprintf buf "@ @[<b 3>sources =%a@]" pp_print_node_set sources;
      fprintf buf "@ @[<hv 3>commands =%a@]" pp_print_command_info_list commands;
      fprintf buf "@]@ }@]"

let pp_print_explicit_rules buf venv =
   fprintf buf "@[<hv 3>Explicit rules:";
   List.iter (fun erule -> fprintf buf "@ %a" pp_print_rule erule) venv.venv_inner.venv_globals.venv_explicit_rules;
   fprintf buf "@]"

(************************************************************************
 * Exceptions
 *)

(*
 * Get the source location for an exception.
 *)
let string_loc = bogus_loc "<Omake_env>"

let rec loc_of_item x =
   match x with
      AstExp e ->
         Omake_ast_util.loc_of_exp e
    | IrExp e ->
         Omake_ir_util.loc_of_exp e
    | Location loc ->
         loc
    | Value _
    | Symbol _
    | String _
    | Error _ ->
         string_loc

(*
 * Value printing.
 *)
let rec pp_print_item buf x =
   match x with
      AstExp e ->
         Omake_ast_print.pp_print_exp buf e

    | IrExp e ->
         Omake_ir_print.pp_print_exp buf e

    | Location _ ->
         ()

    | Symbol v ->
         pp_print_symbol buf v

    | String s ->
         pp_print_string buf s

    | Value v ->
         pp_print_value buf v

    | Error e ->
         pp_print_exn buf e

(*
 * Exception printer.
 *)
and pp_print_exn buf = function
   SyntaxError s ->
      fprintf buf "syntax error: %s" s
 | StringError s ->
      pp_print_string buf s
 | StringIntError (s, i) ->
      fprintf buf "%s: %d" s i
 | StringStringError (s1, s2) ->
      fprintf buf "%s: %s" s1 s2
 | StringVarError (s, v) ->
      fprintf buf "%s: %a" s pp_print_symbol v
 | StringMethodError (s, v) ->
      fprintf buf "%s: %a" s pp_print_method_name v
 | StringNodeError (s, n)->
      fprintf buf "%s: %a" s pp_print_node n
 | StringValueError (s, v) ->
      fprintf buf "@[<hv 3>%s:@ %a@]" s pp_print_value v
 | LazyError printer ->
      printer buf
 | UnboundVar v ->
      fprintf buf "unbound variable: %a" pp_print_symbol v
 | UnboundFun v ->
      fprintf buf "unbound function: %a" pp_print_symbol v
 | UnboundMethod vl ->
      fprintf buf "unbound method: %a" pp_print_method_name vl
 | ArityMismatch (len1, len2) ->
      fprintf buf "arity mismatch: expected %a args, got %d" pp_print_arity len1 len2
 | NotImplemented s ->
      fprintf buf "not implemented: %s" s

(************************************************************************
 * CONSTRUCTION
 ************************************************************************)

module type PosSig =
sig
   val loc_exp_pos    : loc -> pos
   val loc_pos        : loc -> pos -> pos

   val ast_exp_pos    : Omake_ast.exp -> pos
   val ir_exp_pos     : Omake_ir.exp -> pos
   val var_exp_pos    : var -> pos
   val string_exp_pos : string -> pos
   val value_exp_pos  : value -> pos

   val string_pos     : string -> pos -> pos
   val pos_pos        : pos -> pos -> pos
   val int_pos        : int -> pos -> pos
   val var_pos        : var -> pos -> pos
   val error_pos      : omake_error -> pos -> pos

   val del_pos        : (formatter -> unit) -> loc -> pos
   val del_exp_pos    : (formatter -> unit) -> pos -> pos

   (* Utilities *)
   val loc_of_pos     : pos -> loc
   val pp_print_pos   : formatter -> pos -> unit
end

module type NameSig =
sig
   val name : string
end

module MakePos (Name : NameSig) : PosSig =
struct
   module Name' =
   struct
      type t = item

      let name = Name.name

      let loc_of_value = loc_of_item
      let pp_print_value = pp_print_item
   end

   module Pos = Lm_position.MakePos (Name')

   include Pos

   let loc_pos_pos loc pos =
      cons_pos (Location loc) pos

   let ast_exp_pos e    = base_pos (AstExp e)
   let ir_exp_pos e     = base_pos (IrExp e)
   let var_exp_pos v    = base_pos (Symbol v)
   let string_exp_pos s = base_pos (String s)
   let value_exp_pos v  = base_pos (Value v)
   let var_pos          = symbol_pos
   let value_pos v pos  = cons_pos (Value v) pos
   let error_pos e pos  = cons_pos (Error e) pos
end

module Pos = MakePos (struct let name = "Omake_env" end)
open Pos

let () = pp_print_pos_ref := pp_print_pos

(************************************************************************
 * Utilities.
 *)

(*
 * Don't make command info if there are no commands.
 *)
let make_command_info venv sources body =
   match body with
      [] ->
         []
    | _ ->
         [{ command_env     = venv;
            command_sources = sources;
            command_body    = body
          }]

(************************************************************************
 * Channels.
 *)

(*
 * Add a channel slot.
 *)
let venv_add_channel venv file kind mode binary fd =
   let { venv_channel_index = index;
         venv_channels      = channels
       } = venv_runtime
   in
   let data = ChannelValue (Lm_channel.create index file kind mode binary (Some fd)) in
   let channel =
      { channel_id = index;
        channel_data = data
      }
   in
      venv_runtime.venv_channels <- IntTable.add channels index (channel, data);
      venv_runtime.venv_channel_index <- succ index;
      channel

let venv_stdin  = venv_add_channel () "<stdin>"  Lm_channel.PipeChannel Lm_channel.InChannel  false Unix.stdin
let venv_stdout = venv_add_channel () "<stdout>" Lm_channel.PipeChannel Lm_channel.OutChannel false Unix.stdout
let venv_stderr = venv_add_channel () "<stderr>" Lm_channel.PipeChannel Lm_channel.OutChannel false Unix.stderr

(*
 * A formatting channel.
 *)
let venv_add_formatter_channel venv fmt =
   let { venv_channel_index = index;
         venv_channels      = channels
       } = venv_runtime
   in
   let fd = Lm_channel.create index "formatter" Lm_channel.FileChannel Lm_channel.OutChannel true None in
   let data = ChannelValue fd in
   let channel =
      { channel_id = index;
        channel_data = data
      }
   in
   let reader s off len =
      raise (Unix.Unix_error (Unix.EINVAL, "formatter-channel", ""))
   in
   let writer s off len =
      Format.pp_print_string fmt (String.sub s off len);
      len
   in
      Lm_channel.set_io_functions fd reader writer;
      venv_runtime.venv_channels <- IntTable.add channels index (channel, data);
      venv_runtime.venv_channel_index <- succ index;
      channel

(*
 * Get the channel.
 *)
let venv_channel_data channel =
   match channel with
      { channel_data = ChannelDelayed; channel_id = id } ->
         (try
             let _, data = IntTable.find venv_runtime.venv_channels id in
                channel.channel_data <- data;
                data
          with
             Not_found ->
                (* This should never happen, because we close channels on marshaling *)
                raise (Invalid_argument "expand_channel"))
    | { channel_data = data } ->
         data

(*
 * When a channel is closed, close the buffers too.
 *)
let venv_close_channel venv pos channel =
   try
      let () =
         match venv_channel_data channel with
            ChannelClosed ->
               ()
          | ChannelValue channel ->
               Lm_channel.close channel
          | ChannelDelayed ->
               raise (Invalid_argument "venv_close_channel")
      in
         channel.channel_data <- ChannelClosed;
         venv_runtime.venv_channels <- IntTable.remove venv_runtime.venv_channels channel.channel_id
   with
      Not_found ->
         (* Fail silently *)
         ()

(*
 * Get the channel.
 *)
let venv_find_channel venv pos channel =
   let pos = string_pos "venv_find_in_channel" pos in
      match venv_channel_data channel with
         ChannelClosed ->
            raise (OmakeException (pos, StringError "channel is closed"))
       | ChannelValue channel ->
            channel
       | ChannelDelayed ->
            raise (Invalid_argument "venv_find_in_channel")

(*
 * Finding by identifiers.
 *)
let venv_find_channel_id venv pos id =
   let channel, _ =
      try IntTable.find venv_runtime.venv_channels id with
         Not_found ->
            raise (OmakeException (pos, StringIntError ("channel is closed", id)))
   in
      channel

(************************************************************************
 * Primitive values.
 *)

(*
 * Allocate a function primitive.
 *)
let venv_add_prim_fun venv name f =
   let data = Some f in
   let value = { fun_id = name; fun_data = data } in
      venv_runtime.venv_primitives <- SymbolTable.add venv_runtime.venv_primitives name (value, data);
      value

(*
 * Look up the primitive value if we haven't seen it already.
 *)
let venv_apply_prim_fun p venv pos loc args =
   match p with
      { fun_data = Some f } ->
         f venv pos loc args
    | { fun_id = name; fun_data = None } ->
         let opt =
            try
               let _, data = SymbolTable.find venv_runtime.venv_primitives name in
                  p.fun_data <- data;
                  data
            with
               Not_found ->
                  None
         in
            match opt with
               Some f ->
                  f venv pos loc args
             | None ->
                  raise (OmakeException (loc_pos loc pos, UnboundVar name))

(************************************************************************
 * Marshaling.
 *)

(*
 * During marshaling, set all the refs to None.
 *)
let venv_marshal venv f x =
   let { venv_channels = channels;
         venv_primitives = primitives
       } = venv_runtime
   in

   (* Prepare by closing channels, and resetting function data *)
   let prepare () =
      IntTable.iter (fun _ (channel, _) ->
            if channel.channel_id <= 2 then
               channel.channel_data <- ChannelDelayed
            else
               channel.channel_data <- ChannelClosed) channels;
      SymbolTable.iter (fun _ (value, _) ->
            value.fun_data <- None) primitives
   in

   (* Restore by reopening channels and readding function data *)
   let restore () =
      IntTable.iter (fun _ (channel, data) ->
            channel.channel_data <- data) channels;
      SymbolTable.iter (fun _ (value, data) ->
            value.fun_data <- data) primitives
   in
      prepare ();
      try
         let y = f x in
            restore ();
            y
      with
         exn ->
            restore ();
            raise exn

(************************************************************************
 * Methods and objects.
 *)

(*
 * Default empty object.
 *)
let venv_empty_object = SymbolTable.empty

(*
 * Add a class to an object.
 *)
let venv_add_class obj v =
   let table = venv_get_class obj in
   let table = SymbolTable.add table v obj in
      SymbolTable.add obj class_sym (ValClass table)

(*
 * The current object is always in the venv_this field.
 *)
let venv_this venv =
   venv.venv_this

let venv_this_object venv scope =
   match scope with
      ScopeProtected ->
         venv.venv_this
    | ScopePublic ->
         venv.venv_dynamic
    | ScopePrivate ->
         venv.venv_static

let venv_current_object venv classnames =
   let obj = venv.venv_this in
      if classnames = [] then
         obj
      else
         let table = venv_get_class obj in
         let table = List.fold_left (fun table v -> SymbolTable.add table v obj) table classnames in
            SymbolTable.add obj class_sym (ValClass table)

let venv_current_objects venv scope =
   let { venv_this = this;
         venv_dynamic = dynamic;
         venv_static = static
       } = venv
   in
      match scope with
         ScopePrivate ->
            [static]
       | ScopeProtected ->
            [this; dynamic; static]
       | ScopePublic ->
            [dynamic; this; static]

(*
 * Execute a method in an object.
 * If we are currently in the outermost object,
 * push the dynamic scope.
 *)
let venv_with_object venv this =
   { venv with venv_this = this }

(*
 * Define a new object.
 *)
let venv_define_object venv =
   venv_with_object venv SymbolTable.empty

(*
 * Object field lookup.
 *)
let venv_find_field_exn = SymbolTable.find
let venv_add_field      = SymbolTable.add
let venv_object_mem     = SymbolTable.mem
let venv_object_length  = SymbolTable.cardinal
let venv_object_fold    = SymbolTable.fold

let venv_find_field obj pos v =
   try SymbolTable.find obj v with
      Not_found ->
         let pos = string_pos "venv_find_field" pos in
            raise (OmakeException (pos, UnboundVar v))

(*
 * Add the class to the current object.
 *)
let venv_instanceof obj s =
   SymbolTable.mem (venv_get_class obj) s

(*
 * Include the fields in the given class.
 * Be careful to merge classnames.
 *)
let venv_include_object_aux obj1 obj2 =
   let table1 = venv_get_class obj1 in
   let table2 = venv_get_class obj2 in
   let table = SymbolTable.fold SymbolTable.add table1 table2 in
   let obj1 = SymbolTable.fold SymbolTable.add obj1 obj2 in
      SymbolTable.add obj1 class_sym (ValClass table)

let venv_include_object venv obj2 =
   let obj = venv_include_object_aux venv.venv_this obj2 in
      { venv with venv_this = obj }

let venv_flatten_object venv obj2 =
   let obj = venv_include_object_aux venv.venv_dynamic obj2 in
      { venv with venv_dynamic = obj }

(*
 * Get a parent class.
 *)
let venv_find_super venv pos loc v =
   let table = venv_get_class venv.venv_this in
      try SymbolTable.find table v with
         Not_found ->
            let pos = string_pos "venv_find_super" (loc_pos loc pos) in
               raise (OmakeException (pos, StringVarError ("not a super-class", v)))

(*
 * Function scoping.
 *)
let venv_empty_env =
   SymbolTable.empty

let venv_get_env venv =
   venv.venv_static

let venv_with_env venv env =
   { venv with venv_static = env }

(************************************************************************
 * Environment.
 *)

(*
 * Convert a filename to a node.
 *)
let venv_intern venv name =
   Node.intern venv.venv_inner.venv_dir name

let venv_intern_dir venv name =
   Dir.chdir venv.venv_inner.venv_dir name

let venv_intern_list venv names =
   List.map (venv_intern venv) names

let node_set_of_list nodes =
   List.fold_left NodeSet.add NodeSet.empty nodes

let node_set_add_names venv nodes names =
   List.fold_left (fun nodes name ->
         NodeSet.add nodes (venv_intern venv name)) nodes names

let node_set_of_names venv names =
   node_set_add_names venv NodeSet.empty names

(*
 * Compile a wild pattern.
 * It is an error if it isn't wild.
 *)
let compile_wild venv pos loc s =
   if is_wild s then
      wild_compile s
   else
      raise (OmakeException (loc_exp_pos loc, StringStringError ("patterns must be wildcards", s)))

(*
 * Compile a source.
 *)
let compile_source_core venv s =
   if is_wild s then
      SourceWild (wild_compile s)
   else
      SourceNode (venv_intern venv s)

let compile_source venv (kind, s) =
   kind, compile_source_core venv s

(*
 * Perform a wild substitution on a source.
 *)
let subst_source_core venv subst source =
   match source with
      SourceWild wild ->
         let s = wild_subst subst wild in
            venv_intern venv s
    | SourceNode node ->
         node

let subst_source venv subst (kind, source) =
   Node.escape kind (subst_source_core venv subst source)

(*
 * No wildcard matching.
 *)
let intern_source venv (kind, source) =
   Node.escape kind (venv_intern venv source)

(*
 * For variables, try to look them up as 0-arity functions first.
 *)
let venv_find_var_private venv v =
   SymbolTable.find venv.venv_static v

let venv_find_var_protected venv v =
   try SymbolTable.find venv.venv_this v with
      Not_found ->
         try SymbolTable.find venv.venv_dynamic v with
            Not_found ->
               try SymbolTable.find venv.venv_static v with
                  Not_found ->
                     ValString (SymbolTable.find venv.venv_inner.venv_environ v)

let venv_find_var_dynamic venv v =
   try SymbolTable.find venv.venv_dynamic v with
      Not_found ->
         try SymbolTable.find venv.venv_this v with
            Not_found ->
               try SymbolTable.find venv.venv_static v with
                  Not_found ->
                     ValString (SymbolTable.find venv.venv_inner.venv_environ v)

let venv_find_var_exn venv scope v =
   match scope with
      ScopePrivate ->
         venv_find_var_private venv v
    | ScopeProtected ->
         venv_find_var_protected venv v
    | ScopePublic ->
         venv_find_var_dynamic venv v

let venv_get_var venv scope pos v =
   try venv_find_var_exn venv scope v with
      Not_found ->
         let pos = string_pos "venv_get_var" pos in
            raise (OmakeException (pos, UnboundVar v))

let venv_find_var venv scope pos loc v =
   try venv_find_var_exn venv scope v with
      Not_found ->
         let pos = string_pos "venv_find_var" (loc_pos loc pos) in
            raise (OmakeException (loc_pos loc pos, UnboundVar v))

let venv_defined venv scope v =
   List.exists (fun env -> SymbolTable.mem env v) (venv_current_objects venv scope)

(*
 * Adding to variable environment.
 * Add to the current object and the static scope.
 *)
let venv_add_var venv scope pos v s =
   let { venv_this = this;
         venv_static = static;
         venv_dynamic = dynamic
       } = venv
   in
      match scope with
         ScopePrivate ->
            { venv with venv_static  = SymbolTable.add static v s }
       | ScopeProtected ->
            { venv with venv_this    = SymbolTable.add this v s;
                        venv_static  = SymbolTable.add static v s
            }
       | ScopePublic ->
            { venv with venv_dynamic = SymbolTable.add dynamic v s;
                        venv_static  = SymbolTable.add static v s
            }

let venv_add_var_tmp venv v s =
   { venv with venv_dynamic = SymbolTable.add venv.venv_dynamic v s }

let venv_add_var_dynamic venv v s =
   { venv with venv_dynamic = SymbolTable.add venv.venv_dynamic v s;
               venv_static  = SymbolTable.add venv.venv_static v s
   }

let venv_add_assoc venv values =
   List.fold_left (fun venv (v, s) ->
            venv_add_var_dynamic venv v (ValString s)) venv values

let venv_add_args venv pos loc env vl sl =
   let len1 = List.length vl in
   let len2 = List.length sl in
   let _ =
      if len1 <> len2 then
         raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact len1, len2)))
   in
      { venv with venv_dynamic = List.fold_left2 SymbolTable.add venv.venv_dynamic vl sl;
                  venv_static = List.fold_left2 SymbolTable.add env vl sl
      }

(*
 * The system environment.
 *)
let venv_environment venv =
   venv.venv_inner.venv_environ

let venv_getenv venv v =
   SymbolTable.find venv.venv_inner.venv_environ v

let venv_setenv venv v x =
   { venv with venv_inner = { venv.venv_inner with venv_environ = SymbolTable.add venv.venv_inner.venv_environ v x } }

let venv_defined_env venv v =
   SymbolTable.mem venv.venv_inner.venv_environ v

(*
 * Set the options.
 *)
let venv_set_options venv loc pos options =
   let argv = Array.of_list ("omake" :: Lm_string_util.tokens_std options) in
   let add_unknown s =
      raise (OmakeException (loc_pos loc pos, StringStringError ("unknown option", s)))
   in
   let options_spec =
      ["Make options", options_spec]
   in
   let options =
      try Lm_arg.fold_argv argv options_spec venv.venv_inner.venv_options add_unknown "Generic system builder" with
          Lm_arg.BogusArg s ->
            raise (OmakeException (loc_pos loc pos, StringError s))
   in
      { venv with venv_inner = { venv.venv_inner with venv_options = options } }

let venv_options venv =
   venv.venv_inner.venv_options

(************************************************************************
 * Rules
 *)

(*
 * Symbols for directories.
 *)
let wild_sym            = Lm_symbol.add wild_string
let explicit_target_sym = Lm_symbol.add "<EXPLICIT_TARGET>"

(*
 * Don't save explicit rules.
 *)
let venv_explicit_target venv target =
   venv_add_var_dynamic venv explicit_target_sym (ValNode target)

(*
 * Save explicit rules.
 *)
let venv_save_explicit_rules venv loc rules =
   try
      match venv_find_var_dynamic venv explicit_target_sym with
         ValNode target ->
            (* Check that all the rules have this target *)
            List.iter (fun { rule_target = target' } ->
                  if not (Node.equal target' target) then
                     let print_error buf =
                        fprintf buf "@[<hv 3>Computed rule for %a@ has a different target %a@]" (**)
                           pp_print_node target
                           pp_print_node target'
                     in
                        raise (OmakeException (loc_exp_pos loc, LazyError print_error))) rules
       | _ ->
            raise (Invalid_argument "venv_save_explicit_rules")
   with
      Not_found ->
         let globals = venv.venv_inner.venv_globals in
            globals.venv_explicit_new <- List.rev_append rules globals.venv_explicit_new

(*
 * Add an explicit dependency.
 *)
let venv_add_explicit_dep venv loc target source =
   let erule =
      { rule_loc       = loc;
        rule_env       = venv;
        rule_target    = target;
        rule_effects   = NodeSet.singleton target;
        rule_sources   = NodeSet.singleton source;
        rule_match     = None;
        rule_multiple  = false;
        rule_commands  = []
      }
   in
      ignore (venv_save_explicit_rules venv loc [erule])

(*
 * Add the wild target.
 *)
let venv_add_wild_match venv name =
   venv_add_var_tmp venv wild_sym (ValString name)

(*
 * This is the standard way to add results of a pattern match.
 *)
let venv_add_match_args venv args =
   let venv, _ =
      List.fold_left (fun (venv, i) arg ->
            let v = Lm_symbol.add (string_of_int i) in
            let venv = venv_add_var_tmp venv v (ValString arg) in
               venv, succ i) (venv, 1) args
   in
      venv

let venv_add_match venv line args =
   let args = List.map (fun s -> ValString s) args in
   let venv, _ =
      List.fold_left (fun (venv, i) arg ->
            let v = Lm_symbol.add (string_of_int i) in
            let venv = venv_add_var_tmp venv v arg in
               venv, succ i) (venv, 1) args
   in
   let venv = venv_add_var_tmp venv zero_sym (ValString line) in
   let venv = venv_add_var_tmp venv star_sym (ValArray args) in
   let venv = venv_add_var_tmp venv nf_sym   (ValInt (List.length args)) in
      venv

(*
 * Phony names.
 *)
let venv_add_phony venv loc names =
   let inner = venv.venv_inner in
   let { venv_dir = dir;
         venv_phony = phony
       } = inner
   in
   let phony =
      List.fold_left (fun phony name ->
            let gnode = Node.phony_global name in
            let dnode = Node.phony_dir dir name in
            let phony = NodeSet.add phony dnode in
               venv_add_explicit_dep venv loc gnode dnode;
               phony) phony names
   in
      { venv with venv_inner = { inner with venv_phony = phony } }

(*
 * Create an environment.
 *)
let create_environ () =
   let env = Unix.environment () in
   let len = Array.length env in
   let rec collect table i =
      if i = len then
         table
      else
         let s = env.(i) in
         let j = String.index s '=' in
         let v = Lm_symbol.add (String.sub s 0 j) in
         let x = String.sub s (j + 1) (String.length s - j - 1) in
         let table = SymbolTable.add table v x in
            collect table (succ i)
   in
      collect SymbolTable.empty 0

let create options dir exec cache =
   let cwd = Dir.cwd () in
   let env = create_environ () in
   let globals =
      { venv_cwd              = cwd;
        venv_exec             = exec;
        venv_cache            = cache;
        venv_files            = NodeSet.empty;
        venv_directories      = DirTable.empty;
        venv_explicit_rules   = [];
        venv_explicit_new     = [];
        venv_explicit_targets = NodeSet.empty;
        venv_ordering_rules   = [];
        venv_orders           = StringSet.empty;
        venv_pervasives       = SymbolTable.empty
      }
   in
   let inner =
      { venv_dir            = cwd;
        venv_environ        = env;
        venv_phony          = NodeSet.empty;
        venv_no_scanner     = new_symbol_string "<no-scanner>";
        venv_scanner_deps   = [];
        venv_scanner_rules  = [];
        venv_implicit_deps  = [];
        venv_implicit_rules = [];
        venv_globals        = globals;
        venv_options        = options
      }
   in
   let venv =
      { venv_this           = SymbolTable.empty;
        venv_dynamic        = SymbolTable.empty;
        venv_static         = SymbolTable.empty;
        venv_inner          = inner
      }
   in
   let venv = venv_add_phony venv (Lm_location.bogus_loc makeroot_name) [".PHONY"] in
   let venv = venv_add_var_dynamic venv cwd_sym (ValDir cwd) in
   let venv = venv_add_var_dynamic venv stdroot_sym (ValNode (Node.intern Dir.lib "OMakeroot")) in
   let venv = venv_add_var_dynamic venv ostype_sym (ValString Sys.os_type) in
   let venv = venv_add_wild_match venv wild_string in
   let omakepath =
      try
         let path = Lm_string_util.split ":" (Sys.getenv "OMAKEPATH") in
            List.map (venv_intern_dir venv) path
      with
         Not_found ->
            [cwd; Dir.lib]
   in
   let omakepath = ValArray (List.map (fun v -> ValDir v) omakepath) in
   let venv = venv_add_var_dynamic venv omakepath_sym omakepath in
      venv

(*
 * Create a fresh environment from the pervasives.
 * This is used for compiling objects.
 *)
let venv_set_pervasives venv =
   venv.venv_inner.venv_globals.venv_pervasives <- venv.venv_dynamic

let venv_get_pervasives venv node =
   let { venv_inner = inner } = venv in
   let { venv_environ = env;
         venv_options = options;
         venv_globals = globals
       } = inner
   in
   let { venv_cwd = cwd;
         venv_exec = exec;
         venv_cache = cache;
         venv_pervasives = obj
       } = globals
   in
   let globals =
      { venv_cwd              = Node.dir node;
        venv_exec             = exec;
        venv_cache            = cache;
        venv_files            = NodeSet.empty;
        venv_directories      = DirTable.empty;
        venv_explicit_rules   = [];
        venv_explicit_new     = [];
        venv_explicit_targets = NodeSet.empty;
        venv_ordering_rules   = [];
        venv_orders           = StringSet.empty;
        venv_pervasives       = obj
      }
   in
   let inner =
      { venv_dir            = cwd;
        venv_environ        = env;
        venv_phony          = NodeSet.empty;
        venv_no_scanner     = new_symbol_string "<no-scanner>";
        venv_scanner_deps   = [];
        venv_scanner_rules  = [];
        venv_implicit_deps  = [];
        venv_implicit_rules = [];
        venv_globals        = globals;
        venv_options        = options
      }
   in
      { venv_this           = SymbolTable.empty;
        venv_dynamic        = obj;
        venv_static         = SymbolTable.empty;
        venv_inner          = inner
      }

(*
 * Fork the environment, so that changes really have no effect on the old one.
 * This is primarly used when a thread wants a private copy of the environment.
 *)
let venv_fork venv =
   let inner = venv.venv_inner in
   let globals = inner.venv_globals in
   let globals = { globals with venv_cwd = globals.venv_cwd } in
   let inner = { inner with venv_globals = globals } in
      { venv with venv_inner = inner }

(*
 * Global state.
 *)
let venv_exec venv =
   venv.venv_inner.venv_globals.venv_exec

let venv_cache venv =
   venv.venv_inner.venv_globals.venv_cache

(*
 * Change directories.  Update the CWD var, and all a default
 * rule for all the phonies.
 *)
let venv_chdir_dir venv dir =
   { venv with venv_inner = { venv.venv_inner with venv_dir = dir } }

let venv_chdir venv loc dir =
   let inner = venv.venv_inner in
   let { venv_dir = cwd;
         venv_phony = phony
       } = inner
   in
   let dir = Dir.chdir cwd dir in
      if Dir.equal dir cwd then
         venv
      else
         let venv = venv_add_var_dynamic venv cwd_sym (ValDir dir) in
         let venv = venv_chdir_dir venv dir in
         let phony =
            NodeSet.fold (fun phony node ->
                  let node' = Node.phony_chdir node dir in
                     venv_add_explicit_dep venv loc node node';
                     NodeSet.add phony node') NodeSet.empty phony
         in
            { venv with venv_inner = { venv.venv_inner with venv_phony = phony } }

(*
 * The public version does not mess whith the phonies.
 *)
let venv_chdir_dir venv dir =
   let cwd = venv.venv_inner.venv_dir in
      if Dir.equal dir cwd then
         venv
      else
         let venv = venv_add_var_dynamic venv cwd_sym (ValDir dir) in
            venv_chdir_dir venv dir

(*
 * Get the dir.
 *)
let venv_dir venv =
   venv.venv_inner.venv_dir

(*
 * When an OMakefile in a dir is read, save the venv
 * to be used for targets that do not have nay explicit target rule.
 *)
let venv_add_dir venv =
   let globals = venv.venv_inner.venv_globals in
      globals.venv_directories <- DirTable.add globals.venv_directories venv.venv_inner.venv_dir venv

(*
 * When a file is read, remember it as a configuration file.
 *)
let venv_add_file venv node =
   let globals = venv.venv_inner.venv_globals in
      globals.venv_files <- NodeSet.add globals.venv_files node;
      venv

(*
 * Get all the configuration files.
 *)
let venv_files venv =
   venv.venv_inner.venv_globals.venv_files

(*
 * Scanner dependencies.
 *)
let venv_add_scanner_deps venv pos loc patterns sources =
   let pos = string_pos "venv_add_scanner_deps" pos in
   let patterns = List.map (compile_wild venv pos loc) patterns in
   let sources = List.map (compile_source venv) sources in
   let snrule =
      { snrule_patterns = patterns;
        snrule_sources  = sources
      }
   in
      { venv with venv_inner = { venv.venv_inner with venv_scanner_deps = snrule :: venv.venv_inner.venv_scanner_deps } }

(*
 * Scanner dependencies.
 *)
let venv_add_scanner_rule venv pos loc patterns sources body =
   let pos = string_pos "venv_add_scanner_rule" pos in
   let patterns = List.map (compile_wild venv pos loc) patterns in
   let sources = List.map (compile_source venv) sources in
   let srule =
      { srule_loc      = loc;
        srule_patterns = patterns;
        srule_sources  = sources;
        srule_body     = body
      }
   in
      { venv with venv_inner = { venv.venv_inner with venv_scanner_rules = srule :: venv.venv_inner.venv_scanner_rules } }

(*
 * Add a scanner.
 *)
let venv_add_scanner_rule venv pos loc patterns sources commands =
   if commands = [] then
      venv_add_scanner_deps venv pos loc patterns sources
   else
      venv_add_scanner_rule venv pos loc patterns sources commands

(*
 * Add a null rule.
 *)
let venv_add_implicit_deps venv pos loc patterns effects sources =
   let pos = string_pos "venv_add_implicit_deps" pos in
   let () =
      if effects <> [] then
         raise (OmakeException (pos, StringError "implicit dependencies can't have effects"))
   in
   let patterns = List.map (compile_wild venv pos loc) patterns in
   let sources = List.map (compile_source venv) sources in
   let nrule =
      { inrule_loc = loc;
        inrule_patterns = patterns;
        inrule_sources  = sources
      }
   in
   let venv = { venv with venv_inner = { venv.venv_inner with venv_implicit_deps = nrule :: venv.venv_inner.venv_implicit_deps } } in
      venv, []

(*
 * Add an implicit rule.
 *)
let venv_add_implicit_rule venv pos loc patterns effects sources body =
   let pos = string_pos "venv_add_implicit_rule" pos in
   let patterns = List.map (compile_wild venv pos loc) patterns in
   let effects = List.map (compile_source_core venv) effects in
   let sources = List.map (compile_source venv) sources in
   let irule =
      { irule_loc      = loc;
        irule_patterns = patterns;
        irule_effects  = effects;
        irule_sources  = sources;
        irule_body     = body
      }
   in
   let venv = { venv with venv_inner = { venv.venv_inner with venv_implicit_rules = irule :: venv.venv_inner.venv_implicit_rules } } in
      venv, []

(*
 * Add a 2-place explicit rule.
 *)
let venv_add_explicit2_rules venv pos loc multiple targets effects sources body =
   let pos = string_pos "venv_add_explicit_rule" pos in
   let target_args = List.map (venv_intern venv) targets in
   let source_args = List.map (intern_source venv) sources in
   let effects = node_set_add_names venv (node_set_of_list target_args) effects in
   let sources = node_set_of_list source_args in
   let commands = make_command_info venv source_args body in
   let add_target target =
      { rule_loc       = loc;
        rule_env       = venv;
        rule_target    = target;
        rule_effects   = effects;
        rule_sources   = sources;
        rule_match     = None;
        rule_multiple  = multiple;
        rule_commands  = commands
      }
   in
   let rules = List.map add_target target_args in
      venv_save_explicit_rules venv loc rules;
      venv, rules

(*
 * Add a 3-place rule.  Add a separate rule for each target.
 *)
let venv_add_explicit3_rules venv pos loc multiple targets effects patterns sources body =
   let pos = string_pos "venv_add_explicit_rule" pos in
   let patterns = List.map (compile_wild venv pos loc) patterns in
   let sources = List.map (compile_source venv) sources in
   let effects = node_set_of_names venv effects in
   let target_list = List.map (fun s -> venv_intern venv s, s) targets in
   let add_target rules (target, target_str) =
      let rec search patterns =
         match patterns with
            pattern :: patterns ->
               (match wild_match pattern target_str with
                   Some subst ->
                      let source_args = List.map (subst_source venv subst) sources in
                      let sources     = node_set_of_list source_args in
                      let commands    = make_command_info venv source_args body in
                      let erule =
                         { rule_loc      = loc;
                           rule_env      = venv;
                           rule_target   = target;
                           rule_effects  = NodeSet.add effects target;
                           rule_match    = Some (wild_core subst);
                           rule_sources  = sources;
                           rule_multiple = multiple;
                           rule_commands = commands
                         }
                      in
                         erule :: rules
                 | None ->
                      search patterns)
          | [] ->
               raise (OmakeException (loc_pos loc pos, StringError ("bad match: " ^ target_str)))
      in
         search patterns
   in
   let rules = List.fold_left add_target [] target_list in
      venv_save_explicit_rules venv loc rules;
      venv, rules

(*
 * Figure out what to do based on all the parts.
 * A rule is implicit if it has not targets _or_ if the
 * patterns do not contain a %.
 *)
let venv_add_rule venv pos loc multiple targets effects patterns sources commands =
   let pos = string_pos "venv_add_rule" pos in
      match targets, patterns, commands with
         [], [], _ ->
            raise (OmakeException (loc_exp_pos loc, StringError "invalid null rule"))
       | _, [], [] ->
            if List.exists is_wild targets then
               venv_add_implicit_deps venv pos loc targets effects sources
            else
               venv_add_explicit2_rules venv pos loc multiple targets effects sources commands
       | _, [], _ ->
            if List.exists is_wild targets then
               venv_add_implicit_rule venv pos loc targets effects sources commands
            else
               venv_add_explicit2_rules venv pos loc multiple targets effects sources commands
       | _ ->
            venv_add_explicit3_rules venv pos loc multiple targets effects patterns sources commands

(*
 * Flush the explicit list.
 *)
let venv_explicit_flush venv =
   let globals = venv.venv_inner.venv_globals in
   let { venv_explicit_rules = erules;
         venv_explicit_new   = enew;
         venv_explicit_targets = targets
       } = globals
   in
      if enew <> [] then
         let targets, erules =
            List.fold_left (fun (targets, erules) erule ->
                  let targets = NodeSet.add targets erule.rule_target in
                  let erules = erule :: erules in
                     targets, erules) (targets, erules) (List.rev enew)
         in
            globals.venv_explicit_rules <- erules;
            globals.venv_explicit_new <- [];
            globals.venv_explicit_targets <- targets

(*
 * Check if an explicit rule exists.
 *)
let venv_explicit_exists venv target =
   venv_explicit_flush venv;
   NodeSet.mem venv.venv_inner.venv_globals.venv_explicit_targets target

(*
 * Get the explicit rules.  Build a table indexed by target.
 *)
let venv_explicit_rules venv =
   let add_target table target erule =
      NodeTable.filter_add table target (fun entry ->
            match entry with
               Some erule' ->
                  (*
                   * For .PHONY targets, multiple is ignored.
                   * Otherwise, multiple must be the same for both targets.
                   *)
                  if Node.is_phony target
                     || (erule.rule_multiple && erule'.rule_multiple)
                     || ((not erule.rule_multiple && not erule'.rule_multiple)
                            && (erule.rule_commands = [] || erule'.rule_commands = []))
                  then
                     { erule with rule_commands = erule'.rule_commands @ erule.rule_commands }
                  else
                     raise (OmakeException (loc_exp_pos erule.rule_loc, StringNodeError ("multiple ways to build", target)))
             | None ->
                  erule)
   in
   let add_deps table target sources =
      NodeTable.filter_add table target (fun deps ->
            match deps with
               Some deps ->
                  NodeSet.union deps sources
             | None ->
                  sources)
   in
   let info =
      { explicit_targets     = NodeTable.empty;
        explicit_directories = venv.venv_inner.venv_globals.venv_directories;
        explicit_deps        = NodeTable.empty
      }
   in
      venv_explicit_flush venv;
      List.fold_left (fun info erule ->
            let { explicit_targets = target_table;
                  explicit_deps    = dep_table
                } = info
            in
            let { rule_target   = target;
                  rule_sources  = sources
                } = erule
            in
            let target_table = add_target target_table target erule in
            let dep_table    = add_deps   dep_table target sources in
               { info with explicit_targets = target_table;
                           explicit_deps    = dep_table
               }) info (List.rev venv.venv_inner.venv_globals.venv_explicit_rules)


(*
 * Find all the scanner dependencies listed through implicit rules.
 *)
let venv_find_scanner_deps venv target =
   let dir = venv.venv_inner.venv_dir in
   let name = Node.name dir target in
      List.fold_left (fun names snrule ->
            let { snrule_patterns = patterns;
                  snrule_sources = sources
                } = snrule
            in
            let rec search patterns =
               match patterns with
                  pattern :: patterns ->
                     (match wild_match pattern name with
                         Some subst ->
                            List.fold_left (fun names source ->
                                  let source = subst_source venv subst source in
                                     NodeSet.add names source) names sources
                       | None ->
                            search patterns)
                | [] ->
                     names
            in
               search patterns) NodeSet.empty venv.venv_inner.venv_scanner_deps

(*
 * Find a scanner that is appropriate for the given file.
 * This is slow: we search backwards through all the scanners.
 *)
let venv_find_scanner_rules venv target =
   let dir = venv.venv_inner.venv_dir in
   let name = Node.name dir target in
   let rec search matched scanners =
      match scanners with
         srule :: scanners ->
            let { srule_loc      = loc;
                  srule_patterns = patterns;
                  srule_sources  = sources;
                  srule_body     = body
                } = srule
            in
            let _ =
               if debug debug_scanner then
                  eprintf "@[<hv 3>Considering scanner for %s:%a@ sources:%a@]@." (**)
                     name
                     pp_print_wild_list patterns
                     pp_print_source_list sources
            in
            let rec search_pattern matched patterns =
               match patterns with
                  pattern :: patterns ->
                     (match wild_match pattern name with
                         Some subst ->
                            if debug debug_scanner then
                               eprintf "@[<hv 3>Selected scanner for %s:%a@ sources:%a@]@ @[<hv 3>body:%a@]@." (**)
                                  name
                                  pp_print_wild_list patterns
                                  pp_print_source_list sources
                                  pp_print_commands body;
                            let source_args = List.map (subst_source venv subst) sources in
                            let sources = node_set_of_list source_args in
                            let body = make_command_info venv source_args body in
                            let scanner =
                               { scanner_loc      = loc;
                                 scanner_target   = target;
                                 scanner_sources  = sources;
                                 scanner_match    = wild_core subst;
                                 scanner_body     = body
                               }
                            in
                               search_pattern (scanner :: matched) patterns
                       | None ->
                            search_pattern matched patterns)
                | [] ->
                     matched
            in
               search (search_pattern matched patterns) scanners
       | [] ->
            List.rev matched
   in
      if debug debug_scanner then
         eprintf "@[<hv 3>Considering %d scanners for %s@]@." (**)
            (List.length venv.venv_inner.venv_scanner_rules) name;
      search [] venv.venv_inner.venv_scanner_rules

(*
 * Find all the explicit dependencies listed through null
 * rules.
 *)
let venv_find_implicit_deps venv target =
   let dir = venv.venv_inner.venv_dir in
   let name = Node.name venv.venv_inner.venv_dir target in
      List.fold_left (fun names nrule ->
            let { inrule_patterns = patterns;
                  inrule_sources = sources
                } = nrule
            in
            let rec search patterns =
               match patterns with
                  pattern :: patterns ->
                     (match wild_match pattern name with
                         Some subst ->
                            List.fold_left (fun names source ->
                                  let source = subst_source venv subst source in
                                     NodeSet.add names source) names sources
                       | None ->
                            search patterns)
                | [] ->
                     names
            in
               search patterns) NodeSet.empty venv.venv_inner.venv_implicit_deps

(*
 * Find the commands from implicit rules.
 *)
let venv_find_implicit_rules venv target =
   let dir = venv.venv_inner.venv_dir in
   let name = Node.name dir target in
   let _ =
      if debug debug_implicit then
         eprintf "Finding implicit rules for %s@." name
   in
   let rec collect matched = function
      irule :: irules ->
         let _ =
            if debug debug_implicit then
               eprintf "@[<hv 3>venv_find_implicit_rules: considering implicit rule for@ target = %s:@ @[<b 3>patterns =%a@]@ @[<b 3>sources =%a@]@]@." (**)
                  name
                  pp_print_wild_list irule.irule_patterns
                  pp_print_source_list irule.irule_sources
         in
         let rec search = function
            pattern :: patterns ->
               let subst = wild_match pattern name in
                  (match subst with
                      Some _ -> subst
                    | None -> search patterns)
          | [] ->
               None
         in
         let matched =
            match search irule.irule_patterns with
               Some subst ->
                  let source_args = List.map (subst_source venv subst) irule.irule_sources in
                  let commands = make_command_info venv source_args irule.irule_body in
                  let sources = node_set_of_list source_args in
                  let effects =
                     List.fold_left (fun effects pattern ->
                           let effect = wild_subst subst pattern in
                           let effect = venv_intern venv effect in
                              NodeSet.add effects effect) NodeSet.empty irule.irule_patterns
                  in
                  let erule =
                     { rule_loc         = irule.irule_loc;
                       rule_env         = venv;
                       rule_target      = target;
                       rule_match       = Some (wild_core subst);
                       rule_effects     = effects;
                       rule_sources     = sources;
                       rule_multiple    = false;
                       rule_commands    = commands
                     }
                  in
                     if debug debug_implicit then
                        eprintf "@[<hv 3>Added implicit rule for %s:%a@]@." (**)
                           name pp_print_command_info_list commands;
                     erule :: matched
             | None ->
                  matched
         in
            collect matched irules
       | [] ->
            List.rev matched
   in
      collect [] venv.venv_inner.venv_implicit_rules

(************************************************************************
 * Ordering rules.
 *)

(*
 * Add an order.
 *)
let venv_add_orders venv names =
   let globals = venv.venv_inner.venv_globals in
   let orders = List.fold_left StringSet.add globals.venv_orders names in
      globals.venv_orders <- orders;
      venv

(*
 * Check for order.
 *)
let venv_is_order venv name =
   StringSet.mem venv.venv_inner.venv_globals.venv_orders name

(*
 * Add an ordering rule.
 *)
let venv_add_ordering_rule venv pos loc name pattern sources =
   let pos = string_pos "venv_add_ordering_deps" pos in
   let pattern = compile_wild venv pos loc pattern in
   let sources = List.map (compile_source_core venv) sources in
   let orule =
      { orule_loc = loc;
        orule_name = name;
        orule_pattern = pattern;
        orule_sources = sources
      }
   in
   let globals = venv.venv_inner.venv_globals in
      globals.venv_ordering_rules <- orule :: globals.venv_ordering_rules;
      venv

(*
 * Get the ordering dependencies for a name.
 *)
let venv_get_ordering_info venv name =
   List.fold_left (fun orules orule ->
         if Lm_symbol.eq orule.orule_name name then
            orule :: orules
         else
            orules) [] venv.venv_inner.venv_globals.venv_ordering_rules

(*
 * Get extra dependencies.
 *)
let venv_get_ordering_deps venv orules deps =
   let step deps =
      NodeSet.fold (fun deps dep ->
            let target_str = Node.name venv.venv_inner.venv_dir dep in
               List.fold_left (fun deps orule ->
                     let { orule_pattern = pattern;
                           orule_sources = sources
                         } = orule
                     in
                        match wild_match pattern target_str with
                           Some subst ->
                              List.fold_left (fun deps source ->
                                    let source = subst_source_core venv subst source in
                                       NodeSet.add deps source) deps sources
                         | None ->
                              deps) deps orules) deps deps
   in
   let rec fixpoint deps =
      let deps' = step deps in
         if NodeSet.cardinal deps' = NodeSet.cardinal deps then
            deps
         else
            fixpoint deps'
   in
      fixpoint deps

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
