(*
 * Rule evaluation.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2003 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)
open Lm_printf

open Lm_glob
open Lm_debug
open Lm_symbol
open Lm_location
open Lm_string_set

open Omake_ir
open Omake_env
open Omake_node
open Omake_eval
open Omake_exec
open Omake_state
open Omake_value
open Omake_symbol
open Omake_command
open Omake_ir_util
open Omake_ir_print
open Omake_node_sig
open Omake_exec_type
open Omake_exec_util
open Omake_cache_type
open Omake_shell_type

module Pos = MakePos (struct let name = "Omake_rule" end)
open Pos

(*
 * Debugging.
 *)
let debug_active_rules =
   create_debug (**)
      { debug_name = "active-rules";
        debug_description = "Display debugging information for computed rules";
        debug_value = false
      }

(*
 * A data structure to keep sequences of command_info.
 * We can append a command to the current info, or
 * add another entirely new info.
 *)
module type CommandSig =
sig
   type t
   type resume

   (* Create a buffer *)
   val create : venv -> Node.t -> string -> NodeSet.t -> NodeSet.t -> t

   (* Projections *)
   val target : t -> Node.t * string

   (* Adding to the buffer *)
   val add_command : t -> command -> t
   val add_effects : t -> NodeSet.t -> t
   val add_deps    : t -> NodeSet.t -> t

   (* Block operations *)
   val enter : t -> venv -> Node.t list -> resume * t
   val resume : t -> resume -> t

   (* Get the final info *)
   val contents : t -> NodeSet.t * NodeSet.t * command_info list
end

module Command : CommandSig =
struct
   (*
    * The command buffer.
    *)
   type t =
      { buf_target   : Node.t;
        buf_core     : string;
        buf_effects  : NodeSet.t;
        buf_deps     : NodeSet.t;

        (* The state that is being collected *)
        buf_env      : venv;
        buf_sources  : Node.t list;
        buf_commands : command list;

        (* The buffers that have already been collected *)
        buf_info     : command_info list
      }

   type resume = venv * Node.t list

   (*
    * Create a new command buffer.
    *)
   let create venv target core effects deps =
      { buf_target   = target;
        buf_core     = core;
        buf_effects  = effects;
        buf_deps     = deps;
        buf_env      = venv;
        buf_sources  = [];
        buf_commands = [];
        buf_info     = []
      }

   (*
    * Projections.
    *)
   let target buf =
      let { buf_target = target;
            buf_core = core
          } = buf
      in
         target, core

   (*
    * Add a command to the buffer.
    *)
   let add_command buf command =
      { buf with buf_commands = command :: buf.buf_commands }

   let add_effects buf effects =
      { buf with buf_effects = NodeSet.union buf.buf_effects effects }

   let add_deps buf deps =
      { buf with buf_deps = NodeSet.union buf.buf_deps deps }

   (*
    * Start a new environment.
    * Return a state that can be used to resume the current environment.
    *)
   let enter buf venv sources =
      let { buf_env      = venv';
            buf_sources  = sources';
            buf_commands = commands';
            buf_info     = info'
          } = buf
      in
      let info =
         if commands' = [] then
            info'
         else
            let info =
               { command_env     = venv';
                 command_sources = sources';
                 command_body    = List.rev commands'
               }
            in
               info :: info'
      in
      let buf =
         { buf with buf_env      = venv;
                    buf_sources  = sources;
                    buf_commands = [];
                    buf_info     = info
         }
      in
      let resume = venv', sources' in
         resume, buf

   let resume buf (venv, sources) =
      snd (enter buf venv sources)

   (*
    * Get the contents.
    *)
   let contents buf =
      let { buf_env      = venv;
            buf_sources  = sources;
            buf_effects  = effects;
            buf_deps     = deps;
            buf_commands = commands;
            buf_info     = info
          } = buf
      in
      let info =
         if commands = [] then
            info
         else
            let info' =
               { command_env     = venv;
                 command_sources = sources;
                 command_body    = List.rev commands
               }
            in
               info' :: info
      in
      let info = List.rev info in
         effects, deps, info
end

(*
 * Find a rule in a list.
 *)
let rec find_rule loc target rules =
   match rules with
      erule :: rules ->
         if Node.equal target erule.rule_target then
            erule
         else
            find_rule loc target rules
    | [] ->
         raise (OmakeException (loc_exp_pos loc, StringNodeError ("computed rule does not match target", target)))

(*
 * Check if there are any computed commands.
 *)
let commands_are_computed commands =
   List.exists (fun { command_body = body } ->
         List.exists (fun command ->
               match command with
                  CommandSection _ ->
                     true
                | CommandValue _ ->
                     false) body) commands

(*
 * Expand a single command.
 *)
let rec expand_command venv pos loc buf command =
   let pos = string_pos "expand_command" pos in
      match command with
         CommandValue _ ->
            Command.add_command buf command
       | CommandSection (arg, e) ->
            if debug debug_active_rules then
               eprintf "@[<hv 3>section %a@ %a@]@." pp_print_value arg pp_print_exp e;

            (* The section should be either a rule or eval case *)
            match Lm_string_util.trim (string_of_value venv pos arg) with
               "" | "rule" ->
                  expand_rule_section venv pos loc buf e
             | "eval" as s ->
                  expand_eval_section venv pos loc buf s e
             | s ->
                  raise (OmakeException (loc_exp_pos loc, StringStringError ("invalid section argument, valid arguments are rule, eval", s)))

and expand_commands venv pos loc buf commands =
   match commands with
      command :: commands ->
         let buf = expand_command venv pos loc buf command in
            expand_commands venv pos loc buf commands
    | [] ->
         buf

(*
 * This section computes a new rule for the target.
 *)
and expand_rule_section venv pos loc buf e =
   let pos = string_pos "expand_rule_section" pos in
   let target, core = Command.target buf in
   let venv = venv_explicit_target venv target in
   let venv = venv_add_wild_match venv core in
   let v = eval venv e in
      if debug debug_active_rules then
         eprintf "@[<v 0>%a@ @[<hv 3>*** omake: rule body returned@ @[<hv 3>%a@]@]@]@." (**)
            pp_print_location loc
            pp_print_value v;

      match v with
         ValRules erules ->
            let erule = find_rule loc target erules in
            let { rule_effects = effects;
                  rule_sources = deps;
                  rule_commands = commands
                } = erule
            in
            let buf = Command.add_effects buf effects in
            let buf = Command.add_deps buf deps in
               expand_command_info_list pos loc buf commands
       | v ->
            eprintf "@[<v 0>%a@ *** omake: section rule: did not compute a rule@]@." (**)
               pp_print_location loc;
            buf

(*
 * This section is to be evaluated when the rule is run.
 * The rule is dependent on all the files that have been read in so far.
 *)
and expand_eval_section venv pos loc buf s e =
   let pos = string_pos "expand_eval_section" pos in
   let buf = Command.add_deps buf (venv_files venv) in
      Command.add_command buf (CommandSection (ValString s, e))

(*
 * Expand a buf_info list.
 *)
and expand_command_info pos loc buf command =
   let { command_env = venv;
         command_sources = sources;
         command_body = commands
       } = command
   in
   let pos = string_pos "expand_buf_info" pos in
   let resume, buf = Command.enter buf venv sources in
   let buf = expand_commands venv pos loc buf commands in
      Command.resume buf resume

and expand_command_info_list pos loc buf commands =
   let pos = string_pos "expand_buf_info_list" pos in
      List.fold_left (expand_command_info pos loc) buf commands

(*
 * Expand a rule, so that the commands are a command list.
 *)
let expand_rule erule =
   let { rule_loc      = loc;
         rule_env      = venv;
         rule_target   = target;
         rule_effects  = effects;
         rule_match    = core;
         rule_sources  = deps;
         rule_commands = commands
       } = erule
   in
      if commands_are_computed commands then
         let core =
            match core with
               Some s ->
                  s
             | None ->
                  Node.name (venv_dir venv) target
         in
         let pos = string_pos "expand_rule" (loc_exp_pos loc) in
         let buf = Command.create venv target core effects deps in
         let buf = expand_command_info_list pos loc buf commands in
         let effects, deps, commands = Command.contents buf in
            if debug debug_active_rules then
               eprintf "@[<hv 3>expand_rule: %a@ @[<b 3>effects =%a@]@ @[<b 3>deps = %a@]@]@." (**)
                  pp_print_node target
                  pp_print_node_set effects
                  pp_print_node_set deps;
            { erule with rule_effects = effects;
                         rule_sources = deps;
                         rule_commands = commands
            }
      else
         erule

(*
 * Expand a scanner, so that command are a value list.
 *)
let expand_scanner venv scanner =
   let { scanner_loc     = loc;
         scanner_target  = target;
         scanner_match   = core;
         scanner_sources = deps;
         scanner_body    = commands
       } = scanner
   in
      if commands_are_computed commands then
         let pos = string_pos "expand_scanner" (loc_exp_pos loc) in
         let buf = Command.create venv target core NodeSet.empty deps in
         let buf = expand_command_info_list pos loc buf commands in
         let effects, deps, commands = Command.contents buf in

         (* Scanner rules can't have side-effects *)
         let effects = NodeSet.remove effects target in
         let _ =
            if not (NodeSet.is_empty effects) then
               eprintf "@[<v 0>%a@ @[<v 3>*** omake: malformed .SCANNER command has multiple effects@ %a@]@]@." (**)
                  pp_print_location loc
                  pp_print_node_set effects
         in
            { scanner with scanner_sources = deps;
                           scanner_body = commands
            }
      else
         scanner

(************************************************************************
 * Shell utilities.
 *)

(*
 * Get the info for the command.
 *)
let eval_shell_info command =
   let { command_flags = flags;
         command_dir = dir;
         command_target = target
       } = command
   in
      flags, dir, target

(*
 * Kill a process.
 *)
let eval_shell_kill venv pos pid =
   match pid with
      ExternalPid pid ->
         Unix.kill pid Sys.sigterm
    | InternalPid pid ->
         Omake_shell_job.kill venv pos pid SigTerm
    | ResultPid _ ->
         ()

(*
 * Wait for a process to exit.
 *)
let eval_shell_wait venv pos pid =
   let pos = string_pos "eval_shell_wait" pos in
   let _, status, value = Omake_shell_job.waitpid venv pos pid in
      status, value

(************************************************************************
 * Rule evaluation.
 *)

(*
 * From Omake_cache.
 *)
let include_fun = Omake_cache.include_fun
let include_scanner_fun = Omake_cache.include_scanner_fun

(*
 * Glob option parsing.
 *)
let glob_options_of_string options s =
   let len = String.length s in
   let rec search options i =
      if i = len then
         options
      else
         let options =
            match s.[i] with
               'b' -> GlobNoBraces    :: options
             | 'e' -> GlobNoEscape    :: options
             | 'n' -> GlobNoCheck     :: options
             | 'i' -> GlobIgnoreCheck :: options
             | '.' -> GlobDot         :: options
             | 'D' -> GlobOnlyDirs    :: options
             | 'C' -> GlobCVSIgnore   :: options
             | _ -> options
         in
            search options (succ i)
   in
      search options 0

(*
 * Collect the different kinds of sources.
 *)
let add_sources sources kind sources' =
   List.fold_left (fun sources source ->
         (kind, source) :: sources) sources sources'

let sources_of_options venv pos loc sources options =
   let options =
      match eval_value venv pos options with
         ValArray options ->
            options
       | v ->
            raise (OmakeException (loc_pos loc pos, StringValueError ("unknown option argument", v)))
   in
   let effects, sources =
      List.fold_left (fun (effects, sources) option ->
            match option with
               ValArray [optname; optval] ->
                  let s = string_of_value venv pos optname in
                  let v = Lm_symbol.add s in
                  let files = strings_of_value venv pos optval in
                     if Lm_symbol.eq v normal_sym then
                        effects, add_sources sources NodeNormal files
                     else if Lm_symbol.eq v optional_sym then
                        effects, add_sources sources NodeOptional files
                     else if Lm_symbol.eq v exists_sym then
                        effects, add_sources sources NodeExists files
                     else if Lm_symbol.eq v effects_sym then
                        List.rev_append files effects, sources
                     else
                        raise (OmakeException (loc_pos loc pos, StringVarError ("unknown rule option", v)))
             | _ ->
                  raise (OmakeException (loc_pos loc pos, StringValueError ("unknown option value", option)))) ([], sources) options
   in
      List.rev effects, List.rev sources

(*
 * Get the commands.
 *)
let lazy_command venv pos command =
   match command with
      SectionExp (loc, s, el) ->
         CommandSection (eager_string_exp venv pos s, SequenceExp (loc, el))
    | ShellExp (loc, s) ->
         CommandValue (lazy_string_exp venv pos s)
    | _ ->
         CommandSection (ValData "eval", command)

let lazy_commands venv pos commands =
   match commands with
      ValBody (SequenceExp (_, el)) ->
         List.map (lazy_command venv pos) el
    | ValBody e ->
         [lazy_command venv pos e]
    | _ ->
         raise (OmakeException (pos, StringValueError ("unknown rule commands", commands)))

(*
 * Evaluate a rule.  This is the most complicated part of evaluation.
 *
 * There are two types of rules.  Implicit rules are 2-place rules that
 * have a % in the target name.  Explicit rules are 2-place rules
 * that do not have a %, or 3-place rules.
 *
 * In a 3-place rule, the targets are always explicit.
 *)
let rec eval_rule_exp venv pos loc multiple target pattern source options commands =
   let pos = string_pos "eval_rule_exp" pos in

   (* First, evaluate the parts *)
   let targets  = strings_of_value venv pos target in
   let patterns = strings_of_value venv pos pattern in
   let sources  = strings_of_value venv pos source in
   let sources  = add_sources [] NodeNormal sources in
   let effects, sources = sources_of_options venv pos loc sources options in
   let commands = lazy_commands venv pos commands in
   let commands_are_nontrivial = commands <> [] in
      (* Process special rules *)
      match targets with
         [".SUBDIRS"] ->
            if commands_are_nontrivial then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".SUBDIRS rule cannot have build commands"));
            if effects <> [] || patterns <> [] then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".SUBDIRS rule cannot have build patterns or effects"));
            let venv = eval_subdirs_rule venv loc sources in
               ValEnv (venv, ExportAll)
       | [".PHONY"]  ->
            if commands_are_nontrivial then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".PHONY rule cannot have build commands"));
            if effects <> [] || patterns <> [] then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".PHONY rule cannot have build patterns or effects"));
            let sources = List.map snd sources in
            let venv = venv_add_phony venv loc sources in
               ValEnv (venv, ExportAll)
       | [".SCANNER"] ->
            if effects <> [] then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".SCANNER cannot have effects"));
            let targets, sources =
               if patterns = [] then
                  List.map snd sources, []
               else
                  patterns, sources
            in
            let venv = venv_add_scanner_rule venv pos loc targets sources commands in
               ValEnv (venv, ExportAll)
       | [".INCLUDE"] ->
            if effects <> [] then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".INCLUDE cannot have effects"));
            let targets, sources =
               if patterns = [] then
                  List.map snd sources, []
               else
                  patterns, sources
            in
            let venv = eval_include_rule venv pos loc targets sources commands in
               ValEnv (venv, ExportAll)
       | [".ORDER"]  ->
            if commands_are_nontrivial then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".ORDER rule cannot have build commands"));
            if effects <> [] || patterns <> [] then
               raise (OmakeException (loc_exp_pos loc, SyntaxError ".ORDER rule cannot have build patterns or effects"));
            let sources = List.map snd sources in
            let venv = venv_add_phony venv loc sources in
            let venv = venv_add_orders venv sources in
               ValEnv (venv, ExportAll)

         (* .ORDER rules are handled specially *)
       | [name] when venv_is_order venv name ->
            let name = Lm_symbol.add name in
               if commands_are_nontrivial then
                  raise (OmakeException (loc_exp_pos loc, SyntaxError ".ORDER rule cannot have build commands"));
               let venv = eval_ordering_rule venv pos loc name patterns sources in
                  ValEnv (venv, ExportAll)

       | _ ->
            (* Normal rule *)
            let venv, rules = venv_add_rule venv pos loc multiple targets effects patterns sources commands in
               ValEnv (venv, ExportValue (ValRules rules))

(*
 * Read the OMakefiles in the subdirectories too.
 *)
and eval_subdirs_rule venv loc sources =
   List.fold_left (fun venv dir -> eval_subdir venv loc dir) venv sources

(*
 * Compile an OMakefile.
 *)
and eval_subdir venv loc (kind, dir) =
   let _ =
      if kind <> NodeNormal then
         eprintf "*** omake: .SUBDIR kind %a not implemented@." pp_print_node_kind kind
   in
   let cwd = venv_dir venv in
   let venv' = venv_chdir venv loc dir in
   let node = venv_intern venv' makefile_name in
   let venv' = venv_add_file venv' node in
   let name = Node.fullname node in
   let pos = string_pos "eval_subdir" (loc_exp_pos loc) in
   let loc = bogus_loc name in
   let result = eval_include_file venv' pos loc node in

   (*
    * Save the resulting environment as the default to use
    * for targets in this directory.  Also change back to the
    * current directory.
    *)
   let result =
      match result with
         ValEnv (venv, syms) ->
            venv_add_dir venv;
            let venv = venv_chdir_dir venv cwd in
               ValEnv (venv, syms)
       | _ ->
            result
   in
      if debug print_rules then
         eprintf "@[<hv 3>Rules:%a@]@." pp_print_explicit_rules venv;
      fst (add_exports venv pos result)

(*
 * Include all the sources.
 *)
and eval_include_rule venv pos loc sources deps commands =
   let pos = string_pos "eval_include_rule" pos in

   (* Targets and dependencies *)
   let source =
      match sources with
         [source] -> source
       | _ -> raise (OmakeException (pos, StringError ".INCLUDE must have a single source"))
   in
   let target = venv_intern venv source in
   let venv = venv_add_file venv target in
   let deps = List.map (fun (_, dep) -> venv_intern venv dep) deps in

   (* Convert the command list *)
   let commands =
      { Omake_env.command_env = venv;
        Omake_env.command_sources = deps;
        Omake_env.command_body = commands
      }
   in
   let commands = eval_commands venv loc target [commands] in
   let deps = List.fold_left NodeSet.add NodeSet.empty deps in

   (* Ask the cache if this file is up-to-date *)
   let cache = venv_cache venv in
   let updated = Omake_cache.up_to_date cache include_fun deps commands in
   let () =
      (* Run the commands if there are deps, or if the file does not exist *)
      if not updated && commands <> [] && (not (NodeSet.is_empty deps) || Omake_cache.stat cache target = None) then
         exec_commands venv pos target deps commands;

      (* Check that it exists *)
      if Omake_cache.force_stat cache target = None then
         raise (OmakeException (pos, StringNodeError (".INCLUDE rule failed to build the target", target)));

      (* Tell the cache we did the update *)
      if not (NodeSet.is_empty deps) then
         Omake_cache.add cache include_fun (NodeSet.singleton target) deps commands MemoNone
   in
      include_file venv pos loc target

(*
 * Evaluate the commands NOW.
 *)
and waitpid pos exec options pid =
   match Exec.wait exec options with
      WaitExited (id, code, _) ->
         if id <> pid then
            waitpid pos exec options pid
         else if code <> 0 then
            raise (OmakeException (pos, StringIntError ("execution terminated with an error", code)))
    | WaitServer _
    | WaitNotify _
    | WaitNone ->
         waitpid pos exec options pid

and exec_commands venv pos target deps commands =
   let exec = venv_exec venv in
   let cache = venv_cache venv in
   let options = venv_options venv in
      match Exec.spawn exec (eval_shell venv pos) options copy_stdout copy_stderr "include" target commands with
         ProcessFailed ->
            (* The fork failed: we can't continue *)
            raise (OmakeException (pos, StringNodeError ("execution failed", target)))
       | ProcessStarted pid ->
            (* Wait for this job to finish *)
            waitpid pos exec options pid

(*
 * Evaluate the command lines.
 *)
and eval_commands venv loc target commands =
   let rec collect commands' commands =
      match commands with
         command :: commands ->
            let { Omake_env.command_env     = venv;
                  Omake_env.command_sources = sources;
                  Omake_env.command_body    = body
                } = command
            in
            let dir = venv_dir venv in
            let target_arg = Node.name dir target in
            let source_args = List.map (Node.name dir) sources in
            let lines = eval_rule venv loc target_arg source_args body in
            let commands' = List.rev_append lines commands' in
               collect commands' commands
       | [] ->
            List.rev commands'
   in
      collect [] commands

(*
 * Evaluate the rule lines.
 * Add these extra variables.
 *   $@: the target file
 *   $*: the target file, without suffix
 *   $>: the target, without the directory part and  without suffixes
 *   $<: the first source
 *   $+: all the sources
 *   $^: the sources, in alphabetical order, with duplicates removed
 *)
and eval_rule venv loc target sources commands =
   let pos        = string_pos "eval_rule" (loc_exp_pos loc) in
   let root       = Lm_filename_util.root target in
   let root'      = Lm_filename_util.strip_suffixes target in
   let venv       = venv_add_var venv ScopePublic pos star_sym (ValString root) in
   let venv       = venv_add_var venv ScopePublic pos gt_sym   (ValString root') in
   let venv       = venv_add_var venv ScopePublic pos amp_sym  (ValString target) in
   let source_all = String.concat " " sources in
   let source_set = List.fold_left StringSet.add StringSet.empty sources in
   let source_set = StringSet.to_list source_set in
   let source_set = String.concat " " source_set in
   let source =
      match sources with
         source :: _ -> source
       | [] -> ""
   in
   let venv = venv_add_var venv ScopePublic pos plus_sym (ValString source_all) in
   let venv = venv_add_var venv ScopePublic pos hat_sym  (ValString source_set) in
   let venv = venv_add_var venv ScopePublic pos lt_sym   (ValString source) in
   let command_line command =
      match command with
         CommandSection (_, e) ->
            CommandEval e
       | CommandValue v ->
            CommandArgv (strings_of_value venv pos v)
   in
   let commands = List.map command_line commands in
   let dir = venv_dir venv in
   let target = Node.intern dir target in
      parse_commands venv dir target loc commands

(*
 * Add an ordering constraint.
 *)
and eval_ordering_rule venv pos loc name patterns sources =
   let pos = string_pos "eval_ordering_rule" pos in
   let sources = List.map snd sources in
      List.fold_left (fun venv pattern ->
            venv_add_ordering_rule venv pos loc name pattern sources) venv patterns

(************************************************************************
 * Shell.
 *)

(*
 * Get globbing options from the environment.
 *)
and glob_options_of_env venv pos =
   let options = [] in
   let options =
      try
         let s = venv_find_var_exn venv ScopePublic glob_options_sym in
         let s = string_of_value venv pos s in
            glob_options_of_string options s
      with
         Not_found ->
            options
   in
   let options =
      try
         let ignore = venv_find_var_exn venv ScopePublic glob_ignore_sym in
         let ignore = strings_of_value venv pos ignore in
            GlobIgnore ignore :: options
      with
         Not_found ->
            options
   in
   let options =
      try
         let allow = venv_find_var_exn venv ScopePublic glob_allow_sym in
         let allow = strings_of_value venv pos allow in
            GlobAllow allow :: options
      with
         Not_found ->
            options
   in
      options

(*
 * Evaluate a shell expression.
 *)
and eval_shell_exp venv pos loc e =
   let pos  = string_pos "eval_shell_exp" pos in
   let argv = strings_of_value venv pos e in
   let line = Lm_string_util.concat_argv argv in
   let pipe = Omake_shell_lex.parse line in
   let pipe = normalize_pipe venv pos pipe in
   let stdin = channel_of_var venv pos loc stdin_sym in
   let stdout = channel_of_var venv pos loc stdout_sym in
   let stderr = channel_of_var venv pos loc stderr_sym in
   let stdin  = Lm_channel.descr stdin in
   let stdout = Lm_channel.descr stdout in
   let stderr = Lm_channel.descr stderr in
   let result = Omake_shell_job.create_job venv pipe stdin stdout stderr in
      venv, result

(*
 * Construct a shell.  If the SHELL variable is set, then use that shell.
 *)
and eval_shell venv pos =
   let shell =
      { shell_eval        = eval_shell_internal;
        shell_info        = eval_shell_info;
        shell_kill        = eval_shell_kill venv pos;
        shell_wait        = eval_shell_wait venv pos;
        shell_error_value = ValNone;
        shell_print_exp   = pp_print_command_line
      }
   in
      try
         let sh = venv_find_var_exn venv ScopePublic oshell_sym in
         let sh = string_of_value venv pos sh in
            { shell with shell_eval = eval_shell_external sh }
      with
         Not_found ->
            shell

(*
 * Evaluate a shell command using the external shell.
 *)
and eval_shell_external sh stdout stderr command =
   let { command_loc = loc;
         command_dir = dir;
         command_venv = venv;
         command_inst = inst
       } = command
   in
   let pos  = string_pos "eval_shell_external" (loc_exp_pos loc) in
      match inst with
         CommandEval e ->
            eval_command venv stdout stderr pos loc e
       | CommandArgv argv ->
            let line = Lm_string_util.concat_argv argv in
            let line = sprintf "cd %s && (%s)" (Dir.fullname dir) line in
            let pid =
               try Ocaml_patch.create_process sh [|sh; "-c"; line|] Unix.stdin stdout stderr with
                  Unix.Unix_error _ as exn ->
                     raise (UncaughtException (pos, exn))
            in
               ExternalPid pid

(*
 * Evaluate a shell command using the internal shell.
 *)
and eval_shell_internal stdout stderr command =
   let { command_loc = loc;
         command_dir = dir;
         command_venv = venv;
         command_inst = inst
       } = command
   in
   let pos = string_pos "eval_shell_internal" (loc_exp_pos loc) in
      match inst with
         CommandEval e ->
            eval_command venv stdout stderr pos loc e
       | CommandArgv argv ->
            let line = Lm_string_util.concat_argv argv in
            let pipe = Omake_shell_lex.parse line in
            let pipe = normalize_pipe venv pos pipe in
            let pid =
               if !debug_eval then
                  eprintf "eval_shell_internal: creating job@.";
               Omake_shell_job.create_process venv pipe Unix.stdin stdout stderr
            in
               if !debug_eval then
                  eprintf "eval_shell_internal: created job@.";
               pid

(*
 * Used to evaluate expressions.
 *)
and eval_command venv stdout stderr pos loc e =
   let f stdin stdout stderr =
      let venv   = venv_fork venv in
      let stdin  = venv_add_channel venv "<stdin>"  Lm_channel.PipeChannel Lm_channel.InChannel  false stdin in
      let stdout = venv_add_channel venv "<stdout>" Lm_channel.PipeChannel Lm_channel.OutChannel false stdout in
      let stderr = venv_add_channel venv "<stderr>" Lm_channel.PipeChannel Lm_channel.OutChannel false stderr in
      let venv   = venv_add_var venv ScopePublic pos stdin_sym  (ValChannel (InChannel,  stdin)) in
      let venv   = venv_add_var venv ScopePublic pos stdout_sym (ValChannel (OutChannel, stdout)) in
      let venv   = venv_add_var venv ScopePublic pos stderr_sym (ValChannel (OutChannel, stderr)) in
      let () =
         if !debug_eval then
            eprintf "eval_command: evaluating internal function@."
      in
      let code =
         try
            let _ = eval venv e in
               0
         with
            OmakeException _
          | UncaughtException _ as exn ->
               eprintf "%a@." Omake_exn_print.pp_print_exn exn;
               1
          | Unix.Unix_error _
          | Sys_error _
          | Not_found
          | Failure _ as exn ->
               eprintf "%a@." Omake_exn_print.pp_print_exn (UncaughtException (pos, exn));
               1
      in
         if !debug_eval then
            eprintf "eval_command: internal function is done: %d@." code;
         venv_close_channel venv pos stdin;
         venv_close_channel venv pos stdout;
         venv_close_channel venv pos stderr;
         code
   in
      Omake_shell_job.create_thread venv f Unix.stdin stdout stderr

(*
 * Normalize the pipe, so the background is only outermost,
 * and translate commands to aliases.
 *
 * The directory must be an absolute name.
 *)
and normalize_pipe venv pos pipe =
   let pos = string_pos "normalize_pipe" pos in
   let rec normalize squash pipe =
      match pipe with
         PipeApply _ ->
            pipe
       | PipeCommand (loc, command) ->
            normalize_command venv pos loc command
       | PipeCond (loc, op, pipe1, pipe2) ->
            PipeCond (loc, op, normalize true pipe1, normalize true pipe2)
       | PipeCompose (loc, divert_stderr, pipe1, pipe2) ->
            PipeCompose (loc, divert_stderr, normalize true pipe1, normalize true pipe2)
       | PipeBackground (loc, pipe) ->
            let pipe = normalize true pipe in
               if squash then
                  pipe
               else
                  PipeBackground (loc, pipe)
   in
      normalize false pipe

(*
 * Normalize a command.
 * Glob-expand the arguments, and normalize the redirect names.
 *)
and normalize_command venv pos loc command =
   let pos = string_pos "normalize_command" pos in
   let { cmd_argv   = argv;
         cmd_stdin  = stdin;
         cmd_stdout = stdout
       } = command
   in

   (* Do globbing now *)
   let dir = venv_dir venv in
   let argv =
      try glob_argv (glob_options_of_env venv pos) (Dir.fullname dir) argv with
         Failure _ as exn ->
            raise (UncaughtException (pos, exn))
   in

   (* Normalize file names *)
   let stdin =
      match stdin with
         Some stdin ->
            Some (Node.fullname (Node.intern dir stdin))
       | None ->
            None
   in
   let stdout =
      match stdout with
         Some stdout ->
            Some (Node.fullname (Node.intern dir stdout))
       | None ->
            None
   in
   let command =
      { command with cmd_argv = argv;
                     cmd_stdin = stdin;
                     cmd_stdout = stdout
      }
   in
      normalize_apply_command venv pos loc command

(*
 * Normalize a command.
 * Check if it is an internal function first.
 *)
and normalize_apply_command venv pos loc command =
   let pos = string_pos "normalize_apply_command" pos in
      if command.cmd_escaped then
         normalize_shell_command venv pos loc command
      else
         try normalize_apply venv pos loc command with
            Not_found
          | OmakeException _ ->
               normalize_shell_command venv pos loc command

(*
 * This is a normal command, find the full pathname.
 *)
and normalize_shell_command venv pos loc command =
   let pos = string_pos "normalize_shell_command" pos in
   let { cmd_exe = exe } = command in
   let dir = Dir.absname (venv_dir venv) in
   let exe =
      try
         let name = Lm_filename_util.which_dir dir exe in
         let node = Node.intern (Dir.cwd ()) name in
            Node.absname node
      with
         Not_found ->
            raise (OmakeException (pos, StringStringError ("command not found", exe)))
   in
   let command =
      { command with cmd_dir = dir;
                     cmd_exe = exe
      }
   in
      PipeCommand (loc, command)

(*
 * Try to make an internal function.
 * Raises Not_found or OmakeException if not builtin.
 *)
and normalize_apply venv pos loc command =
   let pos = string_pos "normalize_apply" pos in
   let { cmd_env    = env;
         cmd_exe    = exe;
         cmd_argv   = argv;
         cmd_stdin  = stdin;
         cmd_stdout = stdout;
         cmd_stderr = stderr;
         cmd_append = append
       } = command
   in

   (* If this is an internal command, create the PipeApply *)
   let name = Lm_symbol.add exe in
   let f =
      let obj = venv_find_var_exn venv ScopePublic shell_object_sym in
         match eval_single_value venv pos obj with
            ValObject obj ->
               let v = venv_find_field_exn obj name in
               let _, _, f = eval_fun venv pos v in
                  f
          | _ ->
               raise Not_found
   in

   (* Found the function, no exceptions now *)
   let f stdin stdout stderr argv =
      let venv   = venv_fork venv in
      let venv   = List.fold_left (fun venv (v, s) -> venv_setenv venv v s) venv env in
      let stdin  = venv_add_channel venv "<stdin>" Lm_channel.PipeChannel Lm_channel.InChannel  false stdin in
      let stdout = venv_add_channel venv "<stdout>" Lm_channel.PipeChannel Lm_channel.OutChannel false stdout in
      let stderr = venv_add_channel venv "<stderr>" Lm_channel.PipeChannel Lm_channel.OutChannel false stderr in
      let venv   = venv_add_var venv ScopePublic pos stdin_sym  (ValChannel (InChannel,  stdin)) in
      let venv   = venv_add_var venv ScopePublic pos stdout_sym (ValChannel (OutChannel, stdout)) in
      let venv   = venv_add_var venv ScopePublic pos stderr_sym (ValChannel (OutChannel, stderr)) in
      let v      = ValArray (List.map (fun s -> ValString s) argv) in
      let () =
         if !debug_eval then
            eprintf "normalize_apply: evaluating internal function@."
      in
      let code, value =
         try 0, f venv pos loc [v] with
            OmakeException _
          | UncaughtException _ as exn ->
               eprintf "%a@." Omake_exn_print.pp_print_exn exn;
               1, ValNone
          | Unix.Unix_error _
          | Sys_error _
          | Not_found
          | Failure _ as exn ->
               eprintf "%a@." Omake_exn_print.pp_print_exn (UncaughtException (pos, exn));
               1, ValNone
      in
         if !debug_eval then
            eprintf "normalize_apply: internal function is done: %d, %a@." code pp_print_value value;
         venv_close_channel venv pos stdin;
         venv_close_channel venv pos stdout;
         venv_close_channel venv pos stderr;
         if !debug_eval then
            eprintf "normalize_apply: returning value: %d, %a@." code pp_print_value value;
         code, value
   in
   let apply =
      { apply_loc     = loc;
        apply_fun     = f;
        apply_name    = name;
        apply_args    = List.tl argv;
        apply_stdin   = stdin;
        apply_stdout  = stdout;
        apply_stderr  = stderr;
        apply_append  = append
      }
   in
      PipeApply (loc, apply)

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
